from collections import defaultdict
from datetime import date

from .entity import Entity


class Workspace(Entity):

    def __init__(self,
                 name: str,
                 description: str | None = None,
                 created_by: str | None = None) -> None:
        self.name = name
        self.description = description
        self.created_by = created_by

    @property
    def total_time_spent(self) -> int:
        total_time_spent = 0
        for project in self.project:
            total_time_spent += project.total_time_spent
        return total_time_spent

    def daily_time_entries_hours(
            self,
            start_date: str | date | None = None,
            end_date: str | date | None = None,
            last_n_days: int | None = None) -> dict[str, float]:
        entries: dict[str, float] = defaultdict(float)
        for project in self.projects:
            project_entries = project.daily_time_entries_hours(
                start_date, end_date, last_n_days)
            for day, hours in project_entries.items():
                entries[day] += hours
        return entries

    def __str__(self):
        return f"<Workspace {self.id}>: {self.name}"



from __future__ import annotations

from collections import defaultdict
from datetime import datetime
from textual import on, work
from textual.app import App, ComposeResult
from textual.containers import Container, Grid, Horizontal, Vertical
from textual_plotext import PlotextPlot
from textual.reactive import reactive
from textual.screen import ModalScreen, Screen
from textual.validation import Number
from textual.widget import Widget
from textual.widgets import (Button, Input, Header, Footer, Label, Tabs,
                             DataTable, TabbedContent, TabPane, Static,
                             Markdown, MarkdownViewer, Pretty, Rule, Select)

from terka import exceptions
from terka.domain import commands, events, entities
from terka.service_layer import services
from terka.presentations.formatter import Formatter
from terka.presentations.text_ui import components


class SortingMixin:
    current_sorts = set()

    def sort_reverse(self, sort_type: str):
        """Determine if `sort_type` is ascending or descending."""
        reverse = sort_type in self.current_sorts
        if reverse:
            self.current_sorts.remove(sort_type)
        else:
            self.current_sorts.add(sort_type)
        return reverse

    def action_sort(self) -> None:
        table = self.query_one(f"#{self.selected_data_table}", DataTable)
        table.sort(
            self.selected_column,
            reverse=self.sort_reverse(self.selected_column),
        )


class SelectionMixin:

    def on_data_table_cell_selected(self, event: DataTable.CellSelected):
        selected_id = event.cell_key.row_key.value
        self.selected_task = selected_id
        self.selected_column = event.cell_key.column_key.value
        self.selected_data_table = event.data_table.id
        with self.bus.uow as uow:
            if "task" in self.selected_data_table:
                task_obj = uow.tasks.get_by_id(entities.task.Task, selected_id)
                self.query_one(components.Title).text = task_obj.name
                self.query_one(
                    components.Description).text = task_obj.description
                self.query_one(components.Status).value = task_obj.status.name
                self.query_one(
                    components.Priority).value = task_obj.priority.name
                self.query_one(
                    components.Project).value = task_obj.project_.name
                self.query_one(components.Commentaries).values = [
                    (t.date.strftime("%Y-%m-%d %H:%M"), t.text)
                    for t in task_obj.commentaries
                ]
            if "epic" in self.selected_data_table:
                epic_obj = uow.tasks.get_by_id(entities.epic.Epic, selected_id)
                self.query_one(components.Title).text = epic_obj.name
                self.query_one(
                    components.Description).text = epic_obj.description
            if "story" in self.selected_data_table:
                story_obj = uow.tasks.get_by_id(entities.story.Story,
                                                selected_id)
                self.query_one(components.Title).text = story_obj.name
                self.query_one(
                    components.Description).text = story_obj.description


class PopupsMixin:

    def __init__(self):
        ...

    def _process_commands_chain(self,
                                result: tuple[commands.Command, ...],
                                project: str | None = None) -> int | None:
        main_command, *rest = result
        main_command.id = self.selected_task
        if project:
            main_command.project = project
        object_id = self.bus.handle(main_command)
        for cmd in rest:
            cmd.id = object_id or self.selected_task
            if cmd:
                self.bus.handle(cmd)
                self.notify(f"command {cmd}!")
        return object_id

    def action_task_complete(self) -> None:
        self.push_screen(components.TaskComplete(),
                         self.task_complete_callback)

    def task_complete_callback(self, result: tuple[commands.CompleteTask,
                                                   commands.CommentTask,
                                                   commands.TrackTask]):
        self._process_commands_chain(result)
        self.notify(f"Task: {self.selected_task} is completed!")

    def action_task_comment(self) -> None:
        self.push_screen(components.TaskComment(), self.task_comment_callback)

    def task_comment_callback(self, result: commands.CommentTask):
        result.id = self.selected_task
        self.bus.handle(result)
        self.notify(f"Task: {self.selected_task} is commented!")

    def action_task_edit(self) -> None:
        self.push_screen(components.TaskEdit(), self.task_edit_callback)

    def task_edit_callback(self, result: tuple[commands.UpdateTask,
                                               commands.CommentTask]):
        self._process_commands_chain(result)
        self.notify(f"Task: {self.selected_task} is updated!")

    def action_task_add(self) -> None:
        self.push_screen(components.TaskAdd(), self.task_add_callback)

    def task_add_callback(self, result: commands.AddTask):
        if epic := result.epic:
            try:
                self.bus.handle(
                    commands.AddTask(
                        id=self.selected_task,
                        epic=int(epic),
                    ))
                self.notify(
                    f"Task: {self.selected_task} is added to epic {epic}!")
            except exceptions.EntityNotFound:
                self.notify(f"Epic {epic} not found!", severity="error")
        if sprint := result.sprint:
            cmd = commands.AddTask(
                id=self.selected_task,
                sprint=sprint,
            )
            if story_points := result.story_points:
                cmd.story_points = story_points
            try:
                self.bus.handle(cmd)
            except exceptions.EntityNotFound:
                self.notify(f"Sprint {sprint} not found!", severity="error")
            except exceptions.TerkaSprintCompleted:
                self.notify(f"Cannot add task to completed sprint {sprint} !",
                            severity="error")
            if not story_points:
                self.notify(
                    f"Task: {self.selected_task} is added to sprint {sprint}!")
            else:
                self.notify(f"Task: {self.selected_task} story points updated "
                            f"to {story_points}!")
        if story := result.story:
            try:
                self.bus.handle(
                    commands.AddTask(
                        id=self.selected_task,
                        story=story,
                    ))
                self.notify(
                    f"Task: {self.selected_task} is added to story {story}!")
            except exceptions.EntityNotFound:
                self.notify(f"Story {story} not found!", severity="error")

    def action_task_delete(self) -> None:
        if self.selected_column == "due_date":
            self.push_screen(components.TaskDeleteDueDate(),
                             self.task_delete_due_date_callback)
        else:
            self.push_screen(components.TaskDelete(),
                             self.task_delete_callback)

    def task_delete_callback(self, result: tuple[commands.CompleteTask,
                                                 commands.CommentTask,
                                                 commands.TrackTask]):
        delete, *rest = result
        delete.id = self.selected_task
        is_sprint = False
        if hasattr(self, "sprint_id"):
            delete.sprint = self.sprint_id
            is_sprint = True

        self.bus.handle(delete)
        for cmd in rest:
            cmd.id = self.selected_task
            if cmd:
                self.bus.handle(cmd)
                self.notify(f"command {cmd}!")
        if is_sprint:
            self.notify(
                f"Task: {self.selected_task} is deleted from "
                f"sprint {self.sprint_id}!",
                severity="warning")
        else:
            self.notify(f"Task: {self.selected_task} is deleted!",
                        severity="warning")

    def action_task_update_context(self) -> None:
        if self.selected_column == "status":
            self.push_screen(components.TaskStatusEdit(),
                             self.task_update_status_callback)
        if self.selected_column == "priority":
            self.push_screen(components.TaskPriorityEdit(),
                             self.task_update_priority_callback)
        if self.selected_column == "story_points":
            self.push_screen(components.TaskStoryPointsEdit(),
                             self.task_update_story_points_callback)
        if self.selected_column == "time_spent":
            self.push_screen(components.TaskHoursSubmitted(),
                             self.task_update_hours_callback)
        if self.selected_column == "tags":
            self.push_screen(components.TaskTagEdit(),
                             self.task_update_tag_callback)
        if self.selected_column == "collaborators":
            self.push_screen(components.TaskCollaboratorEdit(),
                             self.task_update_collaborator_callback)

    def task_update_status_callback(self, result: str):
        if result:
            self.bus.handle(
                commands.UpdateTask(id=self.selected_task, status=result))
            self.notify(
                f"Task: {self.selected_task} status updated to {result}!")

    def task_update_priority_callback(self, result: str):
        if result:
            self.bus.handle(
                commands.UpdateTask(id=self.selected_task, priority=result))
            self.notify(
                f"Task: {self.selected_task} priority updated to {result}!")

    def task_update_story_points_callback(self, result: str):
        if result:
            self.bus.handle(
                commands.AddTask(id=self.selected_task,
                                 sprint=self.sprint_id,
                                 story_points=result))
            self.notify(
                f"Task: {self.selected_task} story points updated to {result}!"
            )

    def task_update_hours_callback(self, result: str):
        if result:
            self.bus.handle(
                commands.TrackTask(id=self.selected_task, hours=result))
            self.notify(
                f"Tracked {result} minutes for task {self.selected_task}!")

    def task_update_tag_callback(self, result: str):
        if result:
            self.bus.handle(commands.TagTask(id=self.selected_task,
                                             tag=result))
            self.notify(
                f"Tagged task {self.selected_task} with tag(s): {result}!")

    def task_update_collaborator_callback(self, result: str):
        if result:
            self.bus.handle(
                commands.CollaborateTask(id=self.selected_task,
                                         collaborator=result))
            self.notify(
                f"Added collaborator {result} for task {self.selected_task}!")

    def task_delete_due_date_callback(self, result: bool):
        if result:
            self.bus.handle(
                commands.UpdateTask(id=self.selected_task, due_date="Remove"))
            self.notify(f"Removed due date for task {self.selected_task}")

    def action_show_info(self) -> None:
        if "task" in self.selected_data_table:
            sidebar = self.query_one(components.Sidebar)
            # elif "epic" in self.selected_data_table:
            #     sidebar = self.query_one(components.EpicSidebar)
            # elif "story" in self.selected_data_table:
            #     sidebar = self.query_one(components.Sidebar)
            self.set_focus(None)
            if sidebar.has_class("-hidden"):
                sidebar.remove_class("-hidden")
            else:
                if sidebar.query("*:focus"):
                    self.screen.set_focus(None)
                sidebar.add_class("-hidden")


class Comment(Widget):
    value = reactive("text")

    def render(self) -> str:
        return f"Comment: {self.value}"


class TerkaTask(App):
    BINDINGS = [("q", "quit", "Quit")]
    CSS_PATH = "css/vertical_layout.css"

    def __init__(self, entity, bus) -> None:
        super().__init__()
        self.entity = entity
        self.bus = bus

    def compose(self) -> ComposeResult:
        yield Header()
        task_text = f"[bold]Task {self.entity.id}: {self.entity.name}[/bold]"
        if self.entity.is_completed:
            yield Static(task_text, classes="header_completed", id="header")
        elif self.entity.is_overdue:
            yield Static(task_text, classes="header_overdue", id="header")
        else:
            yield Static(task_text, classes="header_simple", id="header")
        yield Static(f"Project: [bold]{self.entity.project_.name}[/bold]",
                     classes="transp")
        yield Static(f"Status: [bold]{self.entity.status.name}[/bold]",
                     classes="transp")
        yield Static(f"Priority: [bold]{self.entity.priority.name}[/bold]",
                     classes="transp")
        created_days_ago = (datetime.now() - self.entity.creation_date).days
        creation_message = f"{created_days_ago} days ago" if created_days_ago > 1 else "today"
        yield Static(f"Created {creation_message}", classes="transp")
        if self.entity.is_completed:
            completion_date = self.entity.completion_date
            if completion_date:
                completed_days_ago = (datetime.now() - completion_date).days
                completion_message = f"{completed_days_ago} days ago" if completed_days_ago > 1 else "today"
                yield Static(
                    f"Completed {completion_message} ({completion_date.strftime('%Y-%m-%d')})",
                    classes="transp")
            else:
                yield Static("Completion date unknown", classes="transp")
        else:
            yield Static(f"Due date: {self.entity.due_date}", classes="transp")
        if sprints := self.entity.sprints:
            sprint_id = ",".join(str(s.sprint) for s in sprints)
            story_points = sprints[-1].story_points
            yield Static(
                f"Sprints: [bold]{sprint_id}[/bold], "
                f"SP: [bold]{story_points}[/bold], "
                f"T: [bold]{Formatter.format_time_spent(self.entity.total_time_spent)}[/bold]",
                classes="transp")
        else:
            yield Static(f"Not in sprint", classes="transp")
        if epics := self.entity.epics:
            epic_ids = ",".join([str(e.epic) for e in epics])
            yield Static(f"Epics: [bold]{epic_ids}[/bold]", classes="transp")
        else:
            yield Static(f"Not in epic", classes="transp")
        if stories := self.entity.stories:
            story_ids = ",".join([str(s.story) for s in stories])
            yield Static(f"Story: [bold]{story_ids}[/bold]", classes="transp")
        else:
            yield Static(f"Not in story", classes="transp")
        if tags := self.entity.tags:
            tags = ",".join(
                t.base_tag.text for t in tags
                if not t.base_tag.text.startswith(("epic", "story", "sprint")))
            yield Static(f"Tags: [bold]{tags}[/bold]", classes="transp")
        else:
            yield Static(f"No tags", classes="transp")
        if collaborators := self.entity.collaborators:
            collaborators = ",".join(c.users.name for c in collaborators)
            yield Static(f"Collaborators: [bold]{collaborators}[/bold]",
                         classes="transp")
        else:
            yield Static(f"No collaborators", classes="transp")
        description_message = self.entity.description or ""
        if description_message:
            yield Static(
                f"[italic]Description...\n[/italic]{description_message}",
                classes="body",
                id="desc")
        else:
            yield Static("[italic]No description...[/italic]",
                         classes="body",
                         id="desc")
        if commentaries := self.entity.commentaries:
            comms = [
                f"[italic]{comment.date.date()}[/italic]: {comment.text}"
                for comment in commentaries
            ]
            comms_message = "\n".join(comms)
            yield Static(f"[italic]Comments...\n\n[/italic]{comms_message}",
                         classes="body",
                         id="history")
        else:
            yield Static("", classes="body", id="history")
        # TODO: implement during next release
        # yield Input(placeholder="Add a comment", classes="body", id="comment")
        # yield Pretty("")

    @on(Input.Submitted)
    def submit_input(self, event: Input.Submitted) -> None:
        self.query_one(Pretty).update("Comment added")
        input = self.query_one(Input)
        input.value = ""


class TerkaProject(App, PopupsMixin, SelectionMixin, SortingMixin):

    CSS_PATH = "css/entities.css"

    BINDINGS = [
        ("n", "new_task", "New Task"),
        ("b", "backlog", "Backlog"),
        ("t", "tasks", "Tasks"),
        ("s", "sort", "Sort"),
        ("S", "sync", "Sync"),
        ("o", "overview", "Overview"),
        ("T", "time", "Time"),
        ("q", "quit", "Quit"),
        ("E", "task_edit", "Edit"),
        ("r", "refresh", "Refresh"),
        ("i", "show_info", "Info"),
        ("a", "task_add", "Add"),
        ("d", "task_complete", "Done"),
        ("C", "task_comment", "Comment"),
        ("U", "task_update_context", "Update"),
        ("X", "task_delete", "Delete"),
        ("Enter", "task_show", "Show"),
    ]

    show_sidebar = reactive(False)

    def __init__(self, entity, bus) -> None:
        super().__init__()
        self.entity = entity
        self.bus = bus
        self.tasks = list()
        self.selected_task = None
        self.selected_column = None
        self.project_id = entity.id
        if _last_synced := self.entity.last_synced:
            self.last_synced = _last_synced[0].sync_date.strftime(
                "%Y-%m-%d %H:%M:%S")
        else:
            self.last_synced = None

    def action_refresh(self):
        self.exit(return_code=4)

    @work(thread=True)
    def action_sync(self) -> None:
        self.notify("syncing project...")
        self.bus.handle(commands.SyncProject(self.project_id))
        self.notify("project synced")

    def on_mount(self) -> None:
        self.title = f"Project: {self.entity.name}"
        self.sub_title = f'Workspace: {self.bus.config.get("workspace")}, last synced: {self.last_synced}'

    def compose(self) -> ComposeResult:
        yield components.Sidebar(classes="-hidden")
        yield Header()
        yield Static(f"{self.entity.name}", classes="header")
        with TabbedContent(initial="tasks"):
            with TabPane("Backlog", id="backlog"):
                yield Button("+Task",
                             id="new_task",
                             variant="success",
                             classes="new_entity")
                table = DataTable(id="project_tasks_backlog_table")
                for column in ("id", "name", "priority", "due_date",
                               "created_at", "assignee", "tags",
                               "collaborators", "time_spent"):
                    table.add_column(column, key=column)
                for task in sorted(self.entity.tasks,
                                   key=lambda x: x.id,
                                   reverse=True):
                    if task.status.name == "BACKLOG":
                        if tags := task.tags:
                            tags_text = ",".join(
                                [tag.base_tag.text for tag in list(tags)])
                        else:
                            tags_text = ""
                        if collaborators := task.collaborators:
                            collaborators_texts = sorted([
                                collaborator.users.name
                                for collaborator in list(collaborators)
                                if collaborator.users
                            ])
                            collaborator_string = ",".join(collaborators_texts)
                        else:
                            collaborator_string = ""
                        if task.is_overdue:
                            task_id = f"[red]{task.id}[/red]"
                        elif task.is_stale:
                            task_id = f"[yellow]{task.id}[/yellow]"
                        else:
                            task_id = str(task.id)
                        if len(commentaries := task.commentaries) > 0:
                            task_name = f"{task.name} [blue][{len(task.commentaries)}][/blue]"
                        else:
                            task_name = task.name
                        table.add_row(task_id,
                                      task_name,
                                      task.priority.name,
                                      task.due_date,
                                      task.creation_date.strftime("%Y-%m-%d"),
                                      str(task.assigned_to.name)
                                      if task.assigned_to else "",
                                      tags_text,
                                      collaborator_string,
                                      Formatter.format_time_spent(
                                          task.total_time_spent),
                                      key=task.id)
                yield table
            with TabPane("Open Tasks", id="tasks"):
                table = DataTable(id="project_open_tasks")
                for column in ("id", "name", "status", "priority", "due_date",
                               "assignee", "tags", "collaborators",
                               "time_spent"):
                    table.add_column(column, key=column)
                for task in sorted(self.entity.tasks,
                                   key=lambda x: x.status.name,
                                   reverse=True):
                    if task.status.name in ("TODO", "IN_PROGRESS", "REVIEW"):
                        if tags := task.tags:
                            tags_text = ",".join(
                                [tag.base_tag.text for tag in list(tags)])
                        else:
                            tags_text = ""
                        if collaborators := task.collaborators:
                            collaborators_texts = sorted([
                                collaborator.users.name
                                for collaborator in list(collaborators)
                                if collaborator.users
                            ])
                            collaborator_string = ",".join(collaborators_texts)
                        else:
                            collaborator_string = ""
                        if task.is_overdue:
                            task_id = f"[red]{task.id}[/red]"
                        elif task.is_stale:
                            task_id = f"[yellow]{task.id}[/yellow]"
                        else:
                            task_id = str(task.id)
                        if len(commentaries := task.commentaries) > 0:
                            task_name = f"{task.name} [blue][{len(task.commentaries)}][/blue]"
                        else:
                            task_name = task.name
                        table.add_row(task_id,
                                      task_name,
                                      task.status.name,
                                      task.priority.name,
                                      task.due_date,
                                      str(task.assigned_to.name)
                                      if task.assigned_to else "",
                                      tags_text,
                                      collaborator_string,
                                      Formatter.format_time_spent(
                                          task.total_time_spent),
                                      key=task.id)
                yield table
            with TabPane("Completed Tasks", id="completed_tasks"):
                table = DataTable(id="project_completed_tasks_table")
                for column in ("id", "name", "status", "priority",
                               "completion_date", "assignee", "tags",
                               "collaborators", "time_spent"):
                    table.add_column(column, key=column)
                completed_tasks = [
                    e for e in self.entity.tasks if e.completion_date
                ]
                for task in sorted(completed_tasks,
                                   key=lambda x: x.completion_date,
                                   reverse=True):
                    if tags := task.tags:
                        tags_text = ",".join(
                            [tag.base_tag.text for tag in list(tags)])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    if len(commentaries := task.commentaries) > 0:
                        task_name = f"{task.name} [blue][{len(task.commentaries)}][/blue]"
                    else:
                        task_name = task.name
                    table.add_row(
                        str(task.id),
                        task_name,
                        task.status.name,
                        task.priority.name,
                        task.completion_date.strftime("%Y-%m-%d"),
                        str(task.assigned_to.name) if task.assigned_to else "",
                        tags_text,
                        collaborator_string,
                        Formatter.format_time_spent(task.total_time_spent),
                        key=task.id)
                yield table
            with TabPane("Epics", id="epics"):
                yield Button("+Epic",
                             id="new_epic",
                             variant="success",
                             classes="new_entity")
                table = DataTable(id="project_epics_table")
                inactive_table = DataTable(id="project_inactive_epics_table")
                for column in ("id", "name", "description", "status",
                               "open_tasks", "completed_tasks"):
                    table.add_column(column, key=f"epic_{column}")
                    inactive_table.add_column(column,
                                              key=f"inactive_epic_{column}")
                for epic in sorted(self.entity.epics,
                                   key=lambda x: len(x.tasks),
                                   reverse=True):
                    if epic.open_tasks:
                        table.add_row(str(epic.id),
                                      shorten_text(epic.name),
                                      shorten_text(epic.description),
                                      epic.status.name,
                                      str(len(epic.open_tasks)),
                                      str(len(epic.completed_tasks)),
                                      key=epic.id)
                    else:
                        inactive_table.add_row(str(epic.id),
                                               shorten_text(epic.name),
                                               shorten_text(epic.description),
                                               epic.status.name,
                                               str(len(epic.open_tasks)),
                                               str(len(epic.completed_tasks)),
                                               key=epic.id)
                if table.row_count:
                    yield Label("Active epics")
                    yield table
                if inactive_table.row_count:
                    yield Rule(line_style="heavy")
                    yield Label("Inactive epics")
                    yield inactive_table
            with TabPane("Stories", id="stories"):
                yield Button("+Story",
                             id="new_story",
                             variant="success",
                             classes="new_entity")
                table = DataTable(id="project_stories_table")
                inactive_table = DataTable(id="project_inactive_stories_table")
                for column in ("id", "name", "description", "status",
                               "open_tasks", "completed_tasks"):
                    table.add_column(column, key=f"story_{column}")
                    inactive_table.add_column(column,
                                              key=f"inactive_story_{column}")
                for story in sorted(self.entity.stories,
                                    key=lambda x: len(x.tasks),
                                    reverse=True):
                    if story.open_tasks:
                        table.add_row(str(story.id),
                                      shorten_text(story.name),
                                      shorten_text(story.description),
                                      story.status.name,
                                      str(len(epic.open_tasks)),
                                      str(len(epic.completed_tasks)),
                                      key=story.id)
                    else:
                        inactive_table.add_row(str(story.id),
                                               shorten_text(story.name),
                                               shorten_text(story.description),
                                               story.status.name,
                                               str(len(epic.open_tasks)),
                                               str(len(epic.completed_tasks)),
                                               key=story.id)
                if table.row_count:
                    yield Label("Active stories")
                    yield table
                if inactive_table.row_count:
                    yield Rule(line_style="heavy")
                    yield Label("Inactive stories")
                    yield inactive_table
            with TabPane("Notes", id="notes"):
                yield Button("+Note",
                             id="new_note",
                             variant="success",
                             classes="new_entity")
                table = DataTable(id="project_notes_table")
                for column in ("id", "text"):
                    table.add_column(column, key=f"note_{column}")
                for note in self.entity.notes:
                    table.add_row(str(note.id), note.name, key=note.id)
                yield table
            with TabPane("Time", id="time"):
                plotext = PlotextPlot(classes="plotext")
                plt = plotext.plt
                n_days = 14
                time = self.entity.daily_time_entries_hours(last_n_days=n_days)
                plt.date_form('Y-m-d')
                plt.title(
                    f"Time tracker - {round(sum(time.values()))} hours spent"
                    f" for the last {n_days} days")
                plt.bar(time.keys(), time.values())
                yield plotext
            with TabPane("Overview", id="overview"):
                collaborators = self.entity.task_collaborators
                sorted_collaborators = ""
                for name, value in sorted(collaborators.items(),
                                          key=lambda x: x[1],
                                          reverse=True):
                    sorted_collaborators += f"  * {name}: {Formatter.format_time_spent(value)} \n"
                yield Markdown(f"""
# Project details:
* Repo: {self.entity.description}
* Time spend: {Formatter.format_time_spent(self.entity.total_time_spent)}
* Collaborators:
{sorted_collaborators}
                """)

        yield Footer()

    def action_epics(self) -> None:
        self.query_one(TabbedContent).active = "epics"

    def action_backlog(self) -> None:
        self.query_one(TabbedContent).active = "backlog"

    def action_tasks(self) -> None:
        self.query_one(TabbedContent).active = "tasks"

    def action_stories(self) -> None:
        self.query_one(TabbedContent).active = "stories"

    def action_overview(self) -> None:
        self.query_one(TabbedContent).active = "overview"

    def action_notes(self) -> None:
        self.query_one(TabbedContent).active = "notes"

    def action_time(self) -> None:
        self.query_one(TabbedContent).active = "time"

    def action_done(self):
        self.selected_task = 0

    def on_data_table_cell_highlighted(self, event: DataTable.CellHighlighted):
        self.selected_task = event.cell_key.row_key.value

    def action_new_task(self):
        self.push_screen(components.NewTask(), self.task_new_callback)

    def action_task_show(self):
        self.push_screen(components.ShowTask())

    @on(Button.Pressed)
    def open_new_element_window(self, event: Button.Pressed) -> None:
        if event.button.id == "new_task":
            self.push_screen(components.NewTask(), self.task_new_callback)
        elif event.button.id == "new_epic":
            self.push_screen(components.NewEpic(), self.epic_new_callback)
        elif event.button.id == "new_story":
            self.push_screen(components.NewStory(), self.story_new_callback)

    def task_new_callback(self, result: list[commands.Command]):
        object_id = self._process_commands_chain(result,
                                                 project=self.project_id)
        self.notify(f"New task created with id {object_id}!")

    def epic_new_callback(self, result: commands.CreateEpic):
        result.project = self.entity.name
        epic = self.bus.handle(result)
        self.notify(f"New epic created: {epic}!")

    def story_new_callback(self, result: commands.CreateStory):
        result.project = self.entity.name
        story = self.bus.handle(result)
        self.notify(f"New story created: {story}!")


class TerkaSprint(App, PopupsMixin, SelectionMixin, SortingMixin):

    CSS_PATH = "css/entities.css"

    BINDINGS = [("n", "new_task", "New Task"), ("t", "tasks", "Tasks"),
                ("N", "notes", "Notes"), ("i", "show_info", "Info"),
                ("o", "overview", "Overview"), ("q", "quit", "Quit"),
                ("s", "sort", "Sort"), ("E", "task_edit", "Edit"),
                ("r", "refresh", "Refresh"), ("d", "task_complete", "Done"),
                ("X", "task_delete", "Delete"), ("a", "task_add", "Add"),
                ("C", "task_comment", "Comment"),
                ("U", "task_update_context", "Update"), ("T", "time", "Time")]

    def __init__(self, entity, bus) -> None:
        super().__init__()
        self.entity = entity
        self.bus = bus
        self.sprint_id = entity.id
        self.tasks = self.get_tasks()
        self.selected_task = None
        self.selected_column = None

    def get_tasks(self, all: bool = False):
        for sprint_task in self.entity.tasks:
            if all:
                yield sprint_task.tasks
            elif sprint_task.tasks.status.name not in ("DONE", "DELETED"):
                yield sprint_task.tasks

    def action_refresh(self):
        self.exit(return_code=4)

    def on_mount(self) -> None:
        self.title = "Sprint"
        self.sub_title = (
            f'Workspace: {self.bus.config.get("workspace")}; '
            f'Time spent today - {Formatter.format_time_spent(self.entity.time_spent_today)}'
        )

    def compose(self) -> ComposeResult:
        yield components.Sidebar(classes="-hidden")
        if self.entity.status.name == "COMPLETED":
            yield Header(classes="completed_sprint")
        elif self.entity.status.name == "ACTIVE":
            yield Header(classes="active_sprint")
        else:
            yield Header(classes="planned_sprint")
        yield Static(
            f"[bold]Sprint {self.entity.id}[/bold]: \n{self.entity.goal}",
            classes="header")
        with TabbedContent(initial="tasks"):
            with TabPane("Tasks", id="tasks"):
                if self.entity.status.name != "COMPLETED":
                    yield Button("+Task",
                                 id="new_task",
                                 variant="success",
                                 classes="new_entity")
                table = DataTable(id="sprint_open_tasks_table")
                for column in ("id", "name", "status", "priority",
                               "story_points", "project", "due_date",
                               "assignee", "tags", "collaborators",
                               "time_spent"):
                    table.add_column(column, key=column)
                for sprint_task in sorted(self.entity.tasks,
                                          key=lambda x: x.tasks.status.value,
                                          reverse=True):
                    task = sprint_task.tasks
                    if tags := task.tags:
                        tags_text = ",".join([
                            tag.base_tag.text for tag in list(tags)
                            if not tag.base_tag.text.startswith("sprint")
                        ])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    if task.is_overdue:
                        task_id = f"[red]{task.id}[/red]"
                    elif task.is_stale:
                        task_id = f"[yellow]{task.id}[/yellow]"
                    else:
                        task_id = str(task.id)
                    if len(commentaries := task.commentaries) > 0:
                        task_name = f"{task.name} [blue][{len(task.commentaries)}][/blue]"
                    else:
                        task_name = task.name
                    if task.status.name not in ("DONE", "DELETED"):
                        table.add_row(task_id,
                                      task_name,
                                      task.status.name,
                                      task.priority.name,
                                      str(sprint_task.story_points),
                                      task.project_.name,
                                      task.due_date,
                                      str(task.assigned_to.name)
                                      if task.assigned_to else "",
                                      tags_text,
                                      collaborator_string,
                                      Formatter.format_time_spent(
                                          task.total_time_spent),
                                      key=task.id)
                yield table
            with TabPane("Completed Tasks", id="completed_tasks"):
                table = DataTable(id="sprint_completed_tasks_table")
                for column in ("id", "name", "project", "completion_date",
                               "assignee", "tags", "collaborators",
                               "story_points", "time_spent"):
                    table.add_column(column, key=column)
                for sprint_task in sorted(self.entity.tasks,
                                          key=lambda x: x.id,
                                          reverse=True):
                    task = sprint_task.tasks
                    if tags := task.tags:
                        tags_text = ",".join([
                            tag.base_tag.text for tag in list(tags)
                            if not tag.base_tag.text.startswith("sprint")
                        ])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    if len(commentaries := task.commentaries) > 0:
                        task_name = f"{task.name} [blue][{len(task.commentaries)}][/blue]"
                    else:
                        task_name = task.name
                    if task.status.name in ("DONE", "DELETED"):
                        table.add_row(
                            str(task.id),
                            task_name,
                            task.project_.name,
                            task.completion_date.strftime("%Y-%m-%d")
                            if task.completion_date else "",
                            str(task.assigned_to.name)
                            if task.assigned_to else "",
                            tags_text,
                            collaborator_string,
                            Formatter.format_time_spent(
                                round(sprint_task.story_points * 60)),
                            Formatter.format_time_spent(task.total_time_spent),
                            key=task.id)
                yield table
            with TabPane("Notes", id="notes"):
                table = DataTable(id="sprint_notes_table")
                table.add_columns("id", "text")
                for task in self.entity.notes:
                    table.add_row(str(task.id), task.name)
                yield table
            with TabPane("Time", id="time"):
                plotext = PlotextPlot(classes="plotext")
                plt = plotext.plt
                if not (sprint_time := self.entity.daily_time_entries_hours()):
                    yield Static("No time tracked")
                else:
                    with self.bus.uow as uow:
                        workspace = services.get_workplace_by_name(
                            self.bus.config.get("workspace"), uow.repo)
                    all_workspace_time = workspace.daily_time_entries_hours(
                        start_date=self.entity.start_date,
                        end_date=self.entity.end_date)
                    non_sprint_times = [
                        all_times - sprint_times
                        for all_times, sprint_times in zip(
                            all_workspace_time.values(), sprint_time.values())
                    ]

                    plt.date_form('Y-m-d')
                    plt.title(
                        f"Time tracker - {Formatter.format_time_spent(self.entity.total_time_spent)} spent"
                    )
                    plt.stacked_bar(sprint_time.keys(),
                                    [sprint_time.values(), non_sprint_times],
                                    label=["sprint", "non-sprint"])
                    yield plotext
            with TabPane("Overview", id="overview"):
                project_split = defaultdict(int)
                for task in self.get_tasks(all=True):
                    project = task.project_.name
                    project_split[project] += task.total_time_spent
                collaborators = self.entity.collaborators
                sorted_collaborators = ""
                for name, value in sorted(collaborators.items(),
                                          key=lambda x: x[1],
                                          reverse=True):
                    sorted_collaborators += f"  * {name}: {Formatter.format_time_spent(value)} \n"
                sorted_projects = ""
                for name, value in sorted(project_split.items(),
                                          key=lambda x: x[1],
                                          reverse=True):
                    sorted_projects += f"  * {name}: {Formatter.format_time_spent(value)} \n"
                yield MarkdownViewer(f"""
# Sprint details:
* Period: {self.entity.start_date} - {self.entity.end_date}
* Open tasks: {len(self.entity.open_tasks)} ({len(self.entity.tasks)})
* Pct Completed: {round(self.entity.pct_completed, 2) :.0%}
* Velocity: {self.entity.velocity}
* Time spend: {Formatter.format_time_spent(self.entity.total_time_spent)}
* Utilization: {round(self.entity.utilization, 2) :.0%}
## Collaborator split:
{sorted_collaborators}
## Project split:
{sorted_projects}
                """)

        yield Footer()

    def action_tasks(self) -> None:
        self.query_one(TabbedContent).active = "tasks"

    def action_time(self) -> None:
        self.query_one(TabbedContent).active = "time"

    def action_overview(self) -> None:
        self.query_one(TabbedContent).active = "overview"

    def action_notes(self) -> None:
        self.query_one(TabbedContent).active = "notes"

    def action_new_task(self):
        if self.entity.status.name != "COMPLETED":
            self.push_screen(components.NewTask(), self.task_new_callback)

    def task_new_callback(self, result: list[commands.Command]):
        object_id = self._process_commands_chain(result)
        self.notify(f"New task created with id {object_id}!")

    @on(Button.Pressed)
    def open_new_element_window(self, event: Button.Pressed) -> None:
        if event.button.id == "new_task":
            self.push_screen(components.NewTask(), self.task_new_callback)


class TerkaComposite(App, PopupsMixin, SelectionMixin, SortingMixin):

    BINDINGS = [("q", "quit", "Quit"), ("r", "refresh", "Refresh")]
    CSS_PATH = "css/entities.css"

    def __init__(self, entity, bus) -> None:
        super().__init__()
        self.entity = entity
        self.bus = bus

    def on_mount(self) -> None:
        self.title = f"Epic: {self.entity.id}"
        self.sub_title = f"Project: {self.entity.project_.name}"

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static(f"{self.entity.name}", classes="header")
        yield Static(f"{self.entity.description}", classes="header")
        with TabbedContent(initial="tasks"):
            with TabPane("Open Tasks", id="tasks"):
                table = DataTable(id="composite_open_tasks")
                for column in ("id", "name", "status", "priority", "due_date",
                               "assignee", "tags", "collaborators",
                               "time_spent"):
                    table.add_column(column, key=column)
                for task in sorted(self.entity.open_tasks,
                                   key=lambda x: x.status.name,
                                   reverse=True):
                    if tags := task.tags:
                        tags_text = ",".join(
                            [tag.base_tag.text for tag in list(tags)])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    if task.is_overdue:
                        task_id = f"[red]{task.id}[/red]"
                    elif task.is_stale:
                        task_id = f"[yellow]{task.id}[/yellow]"
                    else:
                        task_id = str(task.id)
                    if len(commentaries := task.commentaries) > 0:
                        task_name = f"{task.name} [blue][{len(task.commentaries)}][/blue]"
                    else:
                        task_name = task.name
                    table.add_row(
                        task_id,
                        task_name,
                        task.status.name,
                        task.priority.name,
                        task.due_date,
                        str(task.assigned_to.name) if task.assigned_to else "",
                        tags_text,
                        collaborator_string,
                        Formatter.format_time_spent(task.total_time_spent),
                        key=task.id)
                yield table
            with TabPane("Completed Tasks", id="completed_tasks"):
                table = DataTable(id="composite_completed_tasks_table")
                for column in ("id", "name", "status", "priority",
                               "completion_date", "assignee", "tags",
                               "collaborators", "time_spent"):
                    table.add_column(column, key=column)
                for task in sorted(self.entity.completed_tasks,
                                   key=lambda x: x.completion_date,
                                   reverse=True):
                    if tags := task.tags:
                        tags_text = ",".join(
                            [tag.base_tag.text for tag in list(tags)])
                    else:
                        tags_text = ""
                    if collaborators := task.collaborators:
                        collaborators_texts = sorted([
                            collaborator.users.name
                            for collaborator in list(collaborators)
                            if collaborator.users
                        ])
                        collaborator_string = ",".join(collaborators_texts)
                    else:
                        collaborator_string = ""
                    if len(commentaries := task.commentaries) > 0:
                        task_name = f"{task.name} [blue][{len(task.commentaries)}][/blue]"
                    else:
                        task_name = task.name
                    table.add_row(
                        str(task.id),
                        task_name,
                        task.status.name,
                        task.priority.name,
                        task.completion_date.strftime("%Y-%m-%d"),
                        str(task.assigned_to.name) if task.assigned_to else "",
                        tags_text,
                        collaborator_string,
                        Formatter.format_time_spent(task.total_time_spent),
                        key=task.id)
                yield table
            with TabPane("Time", id="time"):
                plotext = PlotextPlot(classes="plotext")
                plt = plotext.plt
                if not (composite_time :=
                        self.entity.daily_time_entries_hours()):
                    yield Static("No time tracked")
                else:
                    plt.date_form('Y-m-d')
                    plt.title(
                        f"Time tracker - {Formatter.format_time_spent(self.entity.total_time_spent)} spent"
                    )
                    plt.stacked_bar(composite_time.keys(),
                                    composite_time.values())
                    yield plotext

    def action_refresh(self):
        self.exit(return_code=4)


class TerkaEpic(TerkaComposite):
    ...


class TerkaStory(TerkaComposite):
    ...


def shorten_text(text: str | None, limit: int = 80) -> str | None:
    if text and len(text) > limit:
        text = text[:limit]
    return text


def get_times(dates: list[str],
              time_entries: list[dict[str, str | int]]) -> list[float]:
    times: list[float] = []
    for date in dates:
        value_for_date = 0
        for entry in time_entries:
            if entry.get("date") == date:
                value_for_date = entry.get("time_spent_hours") / 60
                break
        times.append(value_for_date)
    return times



from typing import Type
from enum import Enum
from dataclasses import dataclass
from collections import defaultdict
from datetime import date, datetime, timedelta
import pandas as pd

from .entity import Entity

class TaskStatus(Enum):
    DELETED = 0
    BACKLOG = 1
    TODO = 2
    IN_PROGRESS = 3
    REVIEW = 4
    DONE = 5


class TaskPriority(Enum):
    UNKNOWN = 0
    LOW = 1
    NORMAL = 2
    HIGH = 3
    URGENT = 4


@dataclass
class TaskInfo:
    name: str
    description: str
    status: str
    priority: str
    due_date: datetime


class Task(Entity):

    def __init__(self,
                 name: str,
                 description: str | None = None,
                 creation_date: datetime = datetime.now(),
                 project: int | None = None,
                 assignee: int | None = None,
                 created_by: int | None = None,
                 due_date: datetime | None = None,
                 status: str = "BACKLOG",
                 priority: str = "NORMAL",
                 sync: bool = True,
                 **kwargs) -> None:
        if not name:
            raise ValueError("task name cannot be empty!")
        if not isinstance(creation_date, datetime):
            raise ValueError(
                "creation_date should be of type datetime.datetime!")
        self.name = name
        self.creation_date = creation_date
        self.description = description
        self.project = project
        self.assignee = assignee
        self.created_by = created_by
        if due_date and due_date.date() < datetime.today().date():
            raise ValueError(
                f"Due date {due_date.date()} cannot be less than today")
        else:
            self.due_date = due_date
        self.status = self._cast_to_enum(TaskStatus, status)
        self.priority = self._cast_to_enum(TaskPriority, priority)
        self.sync = sync

    @property
    def total_time_spent(self):
        if self.time_spent:
            return sum([t.time_spent_minutes for t in self.time_spent])
        return 0

    @property
    def time_spent_today(self):
        if self.time_spent:
            return sum([
                t.time_spent_minutes for t in self.time_spent
                if t.creation_date.date() == datetime.today().date()
            ])
        return 0

    def daily_time_entries_hours(
            self,
            start_date: str | date | None = None,
            end_date: str | date | None = None,
            last_n_days: int | None = None) -> dict[str, float]:
        if last_n_days:
            start_date = (datetime.today() - timedelta(last_n_days)).date()
            end_date = datetime.today().date()
        elif isinstance(start_date, str) and isinstance(end_date, str):
            start_date = datetime.strptime(start_date, "%Y-%m-%d").date()
            end_date = datetime.strptime(end_date, "%Y-%m-%d").date()
        dates = [
            date.strftime("%Y-%m-%d") for date in pd.date_range(
                start_date, end_date).to_pydatetime().tolist()
        ]
        entries = dict.fromkeys(dates, 0.0)
        for entry in self.time_spent:
            creation_date = entry.creation_date.date()
            if creation_date >= start_date and creation_date <= end_date:
                day = creation_date.strftime("%Y-%m-%d")
                entries[day] += entry.time_spent_minutes / 60
        return entries

    @property
    def completion_date(self) -> datetime | None:
        for event in self.history:
            if event.new_value in ("DONE", "DELETED"):
                return event.date

    @property
    def is_stale(self):
        if self.history and self.status.name in ("TODO", "IN_PROGRESS",
                                                 "REVIEW"):
            if max([event.date for event in self.history
                    ]) < (datetime.today() - timedelta(days=5)):
                return True
        return False

    @property
    def is_overdue(self):
        if self.due_date and self.due_date <= date.today():
            return True
        return False

    @property
    def is_completed(self):
        if self.status.name in ("DONE", "DELETED"):
            return True
        return False

    def __repr__(self):
        return f"<Task {self.id}>: {self.name}, {self.status.name}, {self.creation_date}"

    def _cast_to_enum(self, enum: Type[Enum], value: str | Enum) -> Enum:
        return enum[value] if isinstance(value, str) else value

    def __hash__(self):
        return hash(
            ((self.id, self.name, self.description, self.project, self.status,
              self.priority, self.due_date, self.assignee)))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Task):
            return False
        if (self.name, self.description, self.project, self.status,
                self.priority, self.due_date,
                self.assignee) != (other.name, other.description,
                                   other.project, other.status, other.priority,
                                   other.due_date, other.assignee):
            return False
        return True



from __future__ import annotations
from datetime import datetime
from dataclasses import dataclass
import logging
import os

from terka.domain.entities.commentary import TaskCommentary
from terka.domain.entities.task import Task
from terka import exceptions

try:
    import asana
except ImportError:
    raise exceptions.TerkaException(
        "Please install `terka[asana]` to connect to asana")


@dataclass
class AsanaTask:
    id: int
    project: int
    asana_task_id: str
    sync_date: datetime


@dataclass
class AsanaProject:
    id: int
    asana_project_id: str
    sync_date: datetime


@dataclass
class AsanaUser:
    id: int
    asana_user_id: str


class AsanaMigrator:

    status_mapping = {
        "BACKLOG": "Backlog",
        "TODO": "To Do",
        "IN_PROGRESS": "In progress",
        "REVIEW": "Waiting",
        "DONE": "Done",
        "DELETED": "Done"
    }

    reversed_status_mapping = {
        "Backlog": "BACKLOG",
        "To Do": "TODO",
        "In progress": "IN_PROGRESS",
        "Waiting": "Review",
        "Done": "DONE",
    }

    def __init__(
        self,
        client: asana.ApiClient,
        assignee: str | None = os.getenv("ASANA_USER")
    ) -> None:
        self.client = client
        self.assignee = assignee
        self.tasks = asana.TasksApi(client)
        self.sections = asana.SectionsApi(client)
        self.stories = asana.StoriesApi(client)

    def migrate_task(
            self,
            asana_project_id: str,
            task: Task,
            sync_info: dict[str, str] | None = None,
            mapped_external_users: dict[int, str] | None = None) -> str | None:
        try:
            if sync_info:
                asana_task_id = self.update_existing_asana_task(
                    asana_project_id, task, sync_info, mapped_external_users)
            else:
                asana_task_id = self.create_new_asana_task(
                    asana_project_id, task, sync_info, mapped_external_users)
            return asana_task_id
        except Exception as e:
            logging.error(
                f"Failed to migrate task {task.id} to Asana, context: {e}")
            return None

    def update_existing_asana_task(
            self, asana_project_id: str, task: Task, sync_info: dict[str, str],
            mapped_external_users: dict[int, str]) -> str:
        if not (asana_task_id := sync_info.get("asana_task_id")):
            raise exceptions.TerkaException(
                f"Task {task.id} is not sync with Asana")
        opts = {"opt_fields": "name,notes,due_on,assignee"}
        asana_task = self.tasks.get_task(task_gid=asana_task_id, opts=opts)
        update_dict = self._find_task_element_changes(asana_task, task,
                                                      mapped_external_users)
        self.tasks.update_task(task_gid=asana_task_id,
                               body={"data": update_dict},
                               opts=opts)
        self._update_task_status(asana_task_id, task)
        if commentaries := task.commentaries:
            self._migrate_comments_for_task(asana_task_id, commentaries,
                                            sync_info.get("sync_date"))
        logging.info(f"Task {task.id} is updated in Asana")
        return asana_task_id

    def create_new_asana_task(self, asana_project_id: str, task: Task,
                              sync_info: dict,
                              mapped_external_users: dict) -> str:
        asana_task_dict = {
            "name": task.name,
            "notes": task.description or "",
            "projects": [asana_project_id]
        }
        if assignee := task.assignee:
            asana_task_dict["assignee"] = mapped_external_users.get(
                assignee) or self.assignee
        elif self.assignee:
            asana_task_dict["assignee"] = self.assignee
        if due_date := task.due_date:
            asana_task_dict["due_on"] = due_date.strftime("%Y-%m-%d")
        opts = {"opt_fields": "name,projects,due_on,assignee"}
        asana_task = self.tasks.create_task(body={"data": asana_task_dict},
                                            opts=opts)
        asana_task_id = asana_task.get("gid")
        asana_task_status = self.status_mapping[task.status.name]
        self._update_task_status(asana_task_id, task)
        if commentaries := task.commentaries:
            self._migrate_comments_for_task(asana_task_id, commentaries)
        logging.info(f"Task {task.id} is migrated to Asana")
        return asana_task_id

    def _update_task_status(self, asana_task_id: str, task: Task) -> None:
        asana_task_status = self.status_mapping[task.status.name]
        opts = {"body": {"data": {"task": asana_task_id}}}
        self.sections.add_task_for_section(
            section_gid=self.task_statuses[asana_task_status], opts=opts)
        if task.status.name in ("DONE", "DELETED"):
            opts = {"opt_fields": "completed"}
            self.tasks.update_task(task_gid=asana_task_id,
                                   body={"data": {
                                       "completed": True
                                   }},
                                   opts=opts)

    def _migrate_comments_for_task(
            self,
            asana_task_id: str,
            comments: list[TaskCommentary],
            last_sync_date: datetime.datetime | None = None) -> None:
        if last_sync_date:
            comments = [c for c in comments if c.date > last_sync_date]
        else:
            last_sync_date = datetime.fromtimestamp(0)

        opts = {"opt_fields": "text"}
        for comment in comments:
            if comment and comment.date > last_sync_date:
                self.stories.create_story_for_task(
                    task_gid=asana_task_id,
                    body={"data": {
                        "text": comment.text
                    }},
                    opts=opts)

    def _find_task_element_changes(
            self, asana_task: dict, task: Task,
            mapped_external_users: dict) -> dict[str, str]:
        update_dict: dict[str, str] = {}
        if task.name != asana_task.get("name"):
            update_dict["name"] = task.name
        if task.description != asana_task.get("notes", None):
            update_dict["notes"] = task.description or ""
        if asana_assignee := mapped_external_users.get(task.assignee):
            if asana_assignee != asana_task.get("assignee"):
                update_dict["assignee"] = {
                    "gid": asana_assignee,
                    "resource_type": "user"
                }
        if not task.due_date:
            update_dict["due_on"] = None
        else:
            due_date = task.due_date.strftime("%Y-%m-%d")
            if due_date != asana_task.get("due_on"):
                update_dict["due_on"] = due_date
        return update_dict

    def load_task_statuses(self, asana_project_id: str) -> None:
        sections = self.sections.get_sections_for_project(
            asana_project_id, opts={"opt_fields": "name,gid"})
        statuses = {}
        for section in sections:
            statuses[section.get("name")] = section.get("gid")
        self.task_statuses = statuses



import argparse
from datetime import datetime, date
from io import StringIO
import re
import os
from pathlib import Path
from operator import attrgetter
import sys
import yaml
import logging
from pprint import pprint

import rich
from rich.console import Console
from rich.table import Table
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, clear_mappers

from terka.adapters.orm import metadata, start_mappers
from terka.adapters.repository import SqlAlchemyRepository

from terka import bootstrap, exceptions
from terka.domain import commands
from terka.utils import (format_task_dict, process_command, load_config)
from terka.service_layer import handlers, services, unit_of_work

HOME_DIR = os.path.expanduser("~")
DB_URL = f"sqlite:////{HOME_DIR}/.terka/tasks.db"


def init_db(home_dir):
    engine = create_engine(f"sqlite:////{home_dir}/.terka/tasks.db")
    metadata.create_all(engine)
    return engine


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("command", nargs="?")
    parser.add_argument("entity", nargs="?")
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("-v", "--version", dest="version", action="store_true")
    args = parser.parse_known_args()
    args, kwargs = args
    console = Console()
    command, entity = args.command, args.entity
    home_dir = os.path.expanduser("~")
    if args.version:
        import pkg_resources

        version = pkg_resources.require("terka")[0].version
        print(f"terka version {version}")
        exit()
    if args.command == "config":
        console.print(services.get_config())
        exit()
    if args.command == "init":
        services.ServiceCommandHander(home_dir, None,
                                      None).execute(command, None, None)

    file_handler = logging.FileHandler(filename=f"{home_dir}/.terka/terka.log")
    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    log_handlers = [file_handler, stdout_handler]
    logging.basicConfig(
        format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
        handlers=log_handlers,
        level=args.loglevel.upper(),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logger = logging.getLogger(__name__)

    config = load_config(home_dir)
    task_id = config.get("task_id")
    project_name = config.get("project_name")
    if task_id or project_name:
        focus_type = "task" if task_id else "project"
        logger.warning("Using terka in focus mode")
        logger.warning(
            f"Current focus is {focus_type}: {task_id or project_name}")

    task_dict = format_task_dict(config, args, kwargs)
    logger.debug(task_dict)

    service_command_handler = services.ServiceCommandHander(
        home_dir, config, console)
    service_command_handler.execute(command, entity, task_dict)

    bus = bootstrap.bootstrap(start_orm=True,
                              uow=unit_of_work.SqlAlchemyUnitOfWork(DB_URL),
                              config=config)
    queue = []
    queue.append({
        "command": command,
        "entity": entity,
        "task_dict": task_dict
    })
    while queue:
        cmd_dict = queue.pop()
        try:
            handlers.CommandHandler(bus).execute(**cmd_dict)
        except exceptions.TerkaRefreshException:
            queue.append(cmd_dict)


if __name__ == "__main__":
    main()



class Entity:

    def to_dict(self) -> dict:
        result = {}
        for key, value in self.__dict__.items():
            if key.startswith("_"):
                continue
            if hasattr(value, "name"):
                result[key] = value.name
            else:
                result[key] = value
        return result



from collections import defaultdict
from datetime import datetime
from enum import Enum

from .entity import Entity
from .task import Task


class CompositeStatus(Enum):
    ACTIVE = 1
    COMPLETED = 2
    DELETED = 3


class Composite(Entity):

    def __init__(self,
                 name: str,
                 description: str = None,
                 creation_date: datetime = datetime.now(),
                 project: int = None,
                 assignee: int = None,
                 status: str = "ACTIVE",
                 created_by: int = None,
                 **kwargs) -> None:
        if not name:
            raise ValueError("task name cannot be empty!")
        if not isinstance(creation_date, datetime):
            raise ValueError(
                "creation_date should be of type datetime.datetime!")
        self.name = name
        self.creation_date = creation_date
        self.description = description
        self.project = project
        self.assignee = assignee
        self.created_by = created_by
        self.status = status
        self.is_completed = False

    @property
    def backlog_tasks(self) -> list[Task]:
        tasks = []
        for entity_task in self.tasks:
            task = entity_task.tasks
            if task.status.name == "BACKLOG":
                tasks.append(task)
        return tasks
    @property
    def open_tasks(self) -> list[Task]:
        tasks = []
        for entity_task in self.tasks:
            task = entity_task.tasks
            if task.status.name not in ("DONE", "DELETED"):
                tasks.append(task)
        return tasks

    @property
    def completed_tasks(self) -> list[Task]:
        tasks = []
        for entity_task in self.tasks:
            task = entity_task.tasks
            if task.status.name in ("DONE", "DELETED"):
                tasks.append(task)
        return tasks

    def complete(self, tasks) -> None:
        incompleted_tasks = list()
        for task in tasks:
            if task.tasks.status.name != "DONE":
                incompleted_tasks.append(task)
        if incompleted_tasks:
            logging.warning("[composite %d]: %d tasks haven't been completed",
                            self.id, len(incompleted_tasks))
        self.is_completed = True

    def daily_time_entries_hours(self,
                                 last_n_days: int = 14) -> dict[str, float]:
        entries: dict[str, float] = defaultdict(float)
        for task in self.tasks:
            task_entries = task.tasks.daily_time_entries_hours(
                last_n_days=last_n_days)
            for day, hours in task_entries.items():
                entries[day] += hours
        return entries

    @property
    def total_time_spent(self):
        total_time_spent = 0
        for task in self.tasks:
            total_time_spent += task.tasks.total_time_spent
        return total_time_spent



from __future__ import annotations
from typing import Type

from terka.adapters import publisher
from terka.domain import commands, events
from terka.service_layer import unit_of_work, handlers
from terka.presentations.console import printer

Message = commands.Command | events.Event


class MessageBus:

    def __init__(self,
                 uow: unit_of_work.AbstractUnitOfWork,
                 event_handlers: Type[handlers.Handler],
                 command_handlers: Type[handlers.Handler],
                 config: dict | None = None,
                 publisher: publisher.BasePublisher | None = None,
                 ) -> None:
        self.uow = uow
        self.publisher = publisher
        self.event_handlers = event_handlers
        self.command_handlers = command_handlers
        self.config = config
        self.return_value = None
        self.printer = printer.Printer(uow)

    def handle(self, message: Message, context: dict = {}):
        self.queue = [message]
        while self.queue:
            message = self.queue.pop(0)
            if isinstance(message, events.Event):
                self.handle_event(message, context)
            elif isinstance(message, commands.Command):
                self.handle_command(message, context)

        if self.return_value:
            return self.return_value

    def handle_command(self, command: commands.Command,
                       context: dict) -> None:
        handler = self.command_handlers[type(command)]
        if result := handler(command, self, context):
            self.return_value = result
        self.queue.extend(self.uow.collect_new_events())

    def handle_event(self, event: events.Event, context: dict) -> None:
        for handler in self.event_handlers[type(event)]:
            if result := handler(event, self, context):
                self.return_value = result
            self.queue.extend(self.uow.collect_new_events())



from typing import Type
import asana as asn
from collections import defaultdict
from dataclasses import asdict
from datetime import datetime
import functools
import logging
import os
from rich.prompt import Confirm, Prompt

from terka import exceptions, views, utils
from terka.adapters import publisher
from terka.domain import commands, events, entities
from terka.domain.external_connectors import asana
from terka.presentations.vim import templates
from terka.presentations.console import printer
from terka.service_layer import unit_of_work
from terka.utils import create_command

COMMAND_HANDLERS = {}
EVENT_HANDLERS = defaultdict(list)


def register(cmd=None, event=None):

    def fn(func):

        @functools.wraps(func)
        def inner_function(cmd, bus, context):
            return func(cmd, bus, context)

        if cmd:
            COMMAND_HANDLERS[cmd] = inner_function
        elif event:
            EVENT_HANDLERS[event].append(inner_function)
        return inner_function

    return fn


class CommandHandler:

    def __init__(self, bus: "messagebus.MessageBus") -> None:
        self.bus = bus

    def execute(self, command: str, entity: str,
                task_dict: dict | list[dict]) -> None:
        if isinstance(task_dict, list):
            for _task_dict in task_dict:
                self.execute(command, entity, _task_dict)
        else:
            cmd = create_command(command, entity, task_dict)
            try:
                self.bus.handle(cmd, context=task_dict)
            except AttributeError as e:
                print(e)
                raise exceptions.TerkaCommandException(
                    f"Unknown command: `terka {command} {entity}`")


class SprintCommandHandlers:

    @register(cmd=commands.CreateSprint)
    def create(cmd: commands.CreateSprint,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> int:
        if not cmd:
            cmd, context = templates.create_command_from_editor(
                entities.sprint.Sprint, type(cmd))
        with bus.uow as uow:
            # TODO: Add code for checking existing sprints
            new_sprint = entities.sprint.Sprint(**asdict(cmd))
            uow.tasks.add(new_sprint)
            uow.commit()
            new_sprint_id = new_sprint.id
            bus.printer.console.print_new_object(new_sprint)
            return new_sprint_id

    @register(cmd=commands.StartSprint)
    def start(cmd: commands.StartSprint,
              bus: "messagebus.MessageBus",
              context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_sprint := uow.tasks.get_by_id(
                    entities.sprint.Sprint, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Sprint id {cmd.id} is not found")
            if existing_sprint.status.name == "ACTIVE":
                raise exceptions.TerkaSprintActive("Sprint already started")
            if existing_sprint.status.name == "COMPLETED":
                raise exceptions.TerkaSprintCompleted(
                    "Cannot start completed sprint")
            uow.tasks.update(entities.sprint.Sprint, cmd.id,
                             {"status": "ACTIVE"})
            uow.commit()
            for sprint_task in existing_sprint.tasks:
                task = sprint_task.tasks
                task_params = {}
                if task.status.name == "BACKLOG":
                    task_params.update({"status": "TODO"})
                if not task.due_date or task.due_date > existing_sprint.end_date:
                    task_params.update({"due_date": existing_sprint.end_date})
                if task_params:
                    task_params["id"] = task.id
                    uow.published_messages.append(
                        commands.UpdateTask(**task_params))
                # FIXME: ask-input should be provided via CLI
                if sprint_task.story_points == 0 and context.get("ask-input"):
                    story_points = input(
                        "Please enter story points estimation "
                        f"for task <{task.id}>: {task.name}: ")
                    try:
                        story_points = float(story_points)
                        uow.published_messages.append(
                            commands.AddTask(id=task.id,
                                             sprint=existing_sprint.id,
                                             story_points=story_points))
                    except ValueError:
                        print(
                            "[red]Provide number when specifying story points[/red]"
                        )
            logging.debug(f"Sprint started, context: {cmd}")

    @register(cmd=commands.UpdateSprint)
    def update(cmd: commands.UpdateSprint,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_sprint := uow.tasks.get_by_id(
                    entities.sprint.Sprint, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Sprint id {cmd.id} is not found")
            if not cmd:
                cmd, context = templates.create_command_from_editor(
                    existing_sprint, commands.UpdateSprint)
            cmd.inject(bus.config)
            uow.tasks.update(entities.sprint.Sprint, cmd.id,
                             cmd.get_only_set_attributes())
            uow.commit()

    @register(cmd=commands.CompleteSprint)
    def complete(cmd: commands.CompleteSprint,
                 bus: "messagebus.MessageBus",
                 context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_sprint := uow.tasks.get_by_id(
                    entities.sprint.Sprint, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Sprint id {cmd.id} is not found")
            uow.tasks.update(entities.sprint.Sprint, cmd.id,
                             {"status": "COMPLETED"})
            for sprint_task in existing_sprint.tasks:
                task = sprint_task.tasks
                task_params = {}
                if task.status.name == "TODO":
                    task_params.update({"status": "BACKLOG"})
                if task.status.name not in ("DONE",
                                            "DELETED") and task.due_date:
                    task_params.update({"due_date": None})
                if task_params:
                    task_params["id"] = task.id
                    uow.published_messages.append(
                        commands.UpdateTask(**task_params))
            uow.commit()
            logging.debug(f"Sprint completed, context: {cmd}")
        bus.publisher.publish("Topic", events.SprintCompleted(cmd.id))

    @register(cmd=commands.DeleteSprint)
    def delete(cmd: commands.DeleteSprint,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_sprint := uow.tasks.get_by_id(
                    entities.sprint.Sprint, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Sprint id {cmd.id} is not found")
            uow.tasks.update(entities.sprint.Sprint, cmd.id,
                             {"status": "DELETED"})
            uow.commit()
            bus.publisher.publish("Topic", events.SprintDeleted(cmd.id))

    @register(cmd=commands.ShowSprint)
    def show(cmd: commands.ShowSprint,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_sprint := uow.tasks.get_by_id(
                    entities.sprint.Sprint, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Sprint id {cmd.id} is not found")
            bus.printer.tui.print_sprint(existing_sprint, bus)

    @register(cmd=commands.ListSprint)
    def list(cmd: commands.ListSprint,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            if sprints := uow.tasks.list(entities.sprint.Sprint):
                bus.printer.console.print_sprint(
                    sprints, printer.PrintOptions.from_kwargs(**context))


class SprintEventHandlers:

    def assign_story_points(event: events.SprintTaskStoryPointAssigned,
                            bus: "messagebus.MessageBus",
                            context: dict = {}) -> None:
        with bus.uow as uow:
            uow.tasks.update(entities.sprint.SprintTask, event.id,
                             {"story_points": event.story_points})
            uow.commit()
            logging.debug(f"Story point assigned, context: {event}")


class TaskCommandHandlers:

    @register(cmd=commands.CreateTask)
    def create(cmd: commands.CreateTask,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> int:
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                entities.task.Task, commands.CreateTask)
        with bus.uow as uow:
            cmd.inject(bus.config)
            cmd = convert_project(cmd, bus)
            # TODO: make one call
            cmd = convert_user(cmd, bus, "created_by")
            cmd = convert_user(cmd, bus, "assignee")
            new_task = entities.task.Task(**asdict(cmd))
            uow.tasks.add(new_task)
            uow.flush()
            new_task_id = new_task.id
            task_created_event = events.TaskCreated(new_task.id)
            uow.published_messages.append(task_created_event)
            uow.commit()
            TaskCommandHandlers._process_extra_args(new_task.id, context, uow)
            bus.publisher.publish("Topic", task_created_event)
            bus.printer.console.print_new_object(new_task)
            return new_task_id

    @register(cmd=commands.UpdateTask)
    def update(cmd: commands.UpdateTask,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    entities.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            if not cmd:
                cmd, context = templates.create_command_from_editor(
                    existing_task, commands.UpdateTask)
                cmd.id = existing_task.id
            cmd = convert_project(cmd, bus)
            cmd = convert_user(cmd, bus, user_type="assignee")
            for f in cmd.__dataclass_fields__:
                if f == "id":
                    continue
                new_value = getattr(cmd, f)
                old_value = getattr(existing_task, f)
                if hasattr(old_value, "name"):
                    old_value = old_value.name
                if (f == "due_date" and new_value != old_value) or (
                        new_value and new_value != old_value):
                    uow.published_messages.append(
                        events.TaskUpdated(cmd.id,
                                           type=f.upper(),
                                           old_value=old_value,
                                           new_value=new_value))
            update_dict = cmd.get_only_set_attributes()
            if status := update_dict.get("status"):
                update_dict["status"] = entities.task.TaskStatus[status]
            if priority := update_dict.get("priority"):
                update_dict["priority"] = entities.task.TaskPriority[priority]
            uow.tasks.update(entities.task.Task, cmd.id, update_dict)
            uow.commit()
            TaskCommandHandlers._process_extra_args(cmd.id, context, uow)

    @register(cmd=commands.CollaborateTask)
    def collaborate(cmd: commands.CollaborateTask,
                    bus: "messagebus.MessageBus",
                    context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    entities.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            if not (existing_user := uow.tasks.list(
                    entities.user.User, {"name": cmd.collaborator})):
                new_user = entities.user.User(name=cmd.collaborator)
                uow.tasks.add(new_user)
                uow.flush()
                user_id = new_user.id
            else:
                user_id = existing_user[0].id
            if not uow.tasks.list(entities.collaborator.TaskCollaborator, {
                    "task": cmd.id,
                    "collaborator": user_id
            }):
                uow.tasks.add(
                    entities.collaborator.TaskCollaborator(
                        id=cmd.id, collaborator_id=user_id))
                uow.commit()

    @register(cmd=commands.AddTask)
    def add(cmd: commands.AddTask,
            bus: "messagebus.MessageBus",
            context: dict = {}) -> None:

        if (sprint := cmd.sprint) and (story_points := cmd.story_points):
            updated_context = {"entity_type": "sprint", "entity_id": sprint}
            TaskCommandHandlers._add(
                commands.AddTask(cmd.id,
                                 sprint=sprint,
                                 story_points=story_points), bus,
                updated_context)

        else:
            for field in cmd.__dataclass_fields__:
                if field_value := getattr(cmd, field):
                    updated_context = {
                        "entity_type": field,
                        "entity_id": field_value
                    }
                    if field == "sprint":
                        TaskCommandHandlers._add(
                            commands.AddTask(cmd.id, sprint=field_value), bus,
                            updated_context)
                    if field == "epic":
                        TaskCommandHandlers._add(
                            commands.AddTask(cmd.id, epic=field_value), bus,
                            updated_context)
                    if field == "story":
                        TaskCommandHandlers._add(
                            commands.AddTask(cmd.id, story=field_value), bus,
                            updated_context)

    def _add(cmd: commands.AddTask, bus: "messagebus.MessageBus",
             context) -> None:
        entity_name, entity_id = context["entity_type"], context["entity_id"]
        entity_module = getattr(entities, entity_name)
        entity = getattr(entity_module, entity_name.capitalize())
        entity_task_type = getattr(entity_module,
                                   f"{entity_name.capitalize()}Task")
        entity_dict = {"task": cmd.id}
        entity_dict[entity_name] = entity_id
        story_points = cmd.story_points
        with bus.uow as uow:
            if not (existing_entity := uow.tasks.get_by_id(entity, entity_id)):
                raise exceptions.EntityNotFound(
                    f"{entity_name} id {entity_id} is not found")
            if (existing_entity_task := uow.tasks.list(entity_task_type,
                                                       entity_dict)):
                if not story_points:
                    raise exceptions.TaskAddedToEntity(
                        f"task {cmd.id} already added to "
                        f"{entity_name} {entity_id}")
                else:
                    uow.tasks.update(entity_task_type,
                                     existing_entity_task[0].id,
                                     {"story_points": float(story_points)})
                    uow.commit()
            else:
                if entity_name == "sprint" and story_points:
                    entity_dict["story_points"] = story_points
                entity_task = entity_task_type(**entity_dict)
                uow.tasks.add(entity_task)
                uow.commit()
                logging.debug(f"Task added to {entity_name}, context {cmd}")
                if entity_name == "sprint":
                    if existing_entity.status == entities.sprint.SprintStatus.COMPLETED:
                        raise exceptions.TerkaSprintCompleted(
                            f"Sprint {entity_id} is completed")
                    if existing_task := uow.tasks.get_by_id(
                            entities.task.Task, cmd.id):
                        task_params = {}
                        if existing_task.status.name == "BACKLOG":
                            task_params.update({"status": "TODO"})
                        if (not existing_task.due_date
                                or existing_task.due_date >
                                existing_entity.end_date
                                or existing_task.due_date <
                                existing_entity.start_date):
                            task_params.update(
                                {"due_date": existing_entity.end_date})
                        if task_params:
                            task_params["id"] = existing_task.id
                            uow.published_messages.append(
                                commands.UpdateTask(**task_params))

    @register(cmd=commands.AssignTask)
    def assign(cmd: commands.AssignTask,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        ...

    @register(cmd=commands.CompleteTask)
    def complete(cmd: commands.CompleteTask,
                 bus: "messagebus.MessageBus",
                 context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    entities.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            task_completed_event = events.TaskUpdated(
                task=cmd.id,
                type="STATUS",
                old_value=existing_task.status.name,
                new_value="DONE")
            uow.published_messages.append(task_completed_event)
            uow.tasks.update(entities.task.Task, cmd.id,
                             {"status": entities.task.TaskStatus.DONE})
            TaskCommandHandlers._process_extra_args(cmd.id, context, uow)
            uow.commit()
            bus.publisher.publish("Topic", task_completed_event)

    @register(cmd=commands.DeleteTask)
    def delete(cmd: commands.DeleteTask,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:

        updated_context = {}
        for field in cmd.__dataclass_fields__:
            if field == "id":
                continue
            field_value = getattr(cmd, field)

            if field_value := getattr(cmd, field):
                updated_context = {
                    "entity_type": field,
                    "entity_id": field_value
                }
                if field == "sprint":
                    TaskCommandHandlers._delete(
                        commands.DeleteTask(cmd.id, sprint=field_value), bus,
                        updated_context)
                if field == "epic":
                    TaskCommandHandlers._delete(
                        commands.DeleteTask(cmd.id, epic=field_value), bus,
                        updated_context)
                if field == "story":
                    TaskCommandHandlers._delete(
                        commands.DeleteTask(cmd.id, story=field_value), bus,
                        updated_context)
        if not updated_context:
            TaskCommandHandlers._delete(cmd, bus, context)

    def _delete(cmd: commands.DeleteTask,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    entities.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            if "entity_type" in context and "entity_id" in context:
                entity_name, entity_id = context["entity_type"], context[
                    "entity_id"]
                entity_module = getattr(entities, entity_name)
                entity = getattr(entity_module, entity_name.capitalize())
                entity_task_type = getattr(entity_module,
                                           f"{entity_name.capitalize()}Task")
                entity_dict = {"task": cmd.id}
                entity_dict[entity_name] = entity_id
                if not (existing_entity := uow.tasks.get_by_id(
                        entity, entity_id)):
                    raise exceptions.EntityNotFound(
                        f"{entity_name} id {entity_id} is not found")
                if existing_entity_task := uow.tasks.list(
                        entity_task_type, entity_dict):
                    uow.tasks.delete(entity_task_type,
                                     existing_entity_task[0].id)
                    uow.commit()
                    logging.debug(
                        f"Task deleted from {entity_name.capitalize()} "
                        f"{entity_id}, context {cmd}")
                    if entity_name == "sprint":
                        if existing_task := uow.tasks.get_by_id(
                                entities.task.Task, cmd.id):
                            task_params = {}
                            if existing_task.status.name == "TODO":
                                task_params.update({"status": "BACKLOG"})
                            task_params.update({"due_date": None})
                            if task_params:
                                # FIXME: Duplication
                                task_params["id"] = existing_task.id
                                uow.published_messages.append(
                                    commands.UpdateTask(**task_params))
            else:
                task_deleted_event = events.TaskUpdated(
                    task=cmd.id,
                    type="STATUS",
                    old_value=existing_task.status.name,
                    new_value="DELETED")
                uow.published_messages.append(task_deleted_event)
                uow.tasks.update(entities.task.Task, cmd.id,
                                 {"status": entities.task.TaskStatus.DELETED})
                uow.published_messages.append(task_deleted_event)
            TaskCommandHandlers._process_extra_args(cmd.id, context, uow)
            uow.commit()
        bus.publisher.publish("Topic", events.TaskCompleted(cmd.id))

    @register(cmd=commands.CommentTask)
    def comment(cmd: commands.CommentTask,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            if text := cmd.text.strip():
                uow.tasks.add(
                    entities.commentary.TaskCommentary(id=cmd.id, text=text))
                uow.tasks.update(entities.task.Task, cmd.id,
                                 {"modification_date": datetime.now()})
                uow.commit()

    @register(cmd=commands.TrackTask)
    def track(cmd: commands.TrackTask,
              bus: "messagebus.MessageBus",
              context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    entities.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            uow.tasks.add(
                entities.time_tracker.TimeTrackerEntry(
                    task=cmd.id, time_spent_minutes=cmd.hours))
            uow.commit()

    @register(cmd=commands.TagTask)
    def tag(cmd: commands.TagTask,
            bus: "messagebus.MessageBus",
            context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_task := uow.tasks.get_by_id(
                    entities.task.Task, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Task id {cmd.id} is not found")
            if not (existing_tag := uow.tasks.list(entities.tag.BaseTag,
                                                   {"text": cmd.tag})):
                new_tag = entities.tag.BaseTag(text=cmd.tag)
                uow.tasks.add(new_tag)
                uow.flush()
                tag_id = new_tag.id
            else:
                tag_id = existing_tag[0].id
            if not uow.tasks.list(entities.tag.TaskTag, {
                    "task": cmd.id,
                    "tag": tag_id
            }):
                uow.tasks.add(entities.tag.TaskTag(id=cmd.id, tag_id=tag_id))
                uow.commit()

    def _process_extra_args(id, context, uow):
        if tags := context.get("tags"):
            for tag in tags.split(","):
                uow.published_messages.append(commands.TagTask(id=id, tag=tag))
        if collaborators := context.get("collaborators"):
            for collaborator_name in collaborators.split(","):
                uow.published_messages.append(
                    commands.CollaborateTask(id=id,
                                             collaborator=collaborator_name))
        if sprints := context.get("sprint"):
            for sprint in sprints.split(","):
                uow.published_messages.append(
                    commands.AddTask(id=id, sprint=sprint))
        if epics := context.get("epic"):
            for epic in epics.split(","):
                uow.published_messages.append(commands.AddTask(id=id, epic=epic))
        if stories := context.get("story"):
            for story in stories.split(","):
                uow.published_messages.append(
                    commands.AddTask(id=id, story=story))
        if comment := context.get("comment"):
            uow.published_messages.append(
                commands.CommentTask(id=id, text=comment))
        if hours := context.get("time_spent"):
            uow.published_messages.append(commands.TrackTask(id=id, hours=hours))

    @register(cmd=commands.ListTask)
    def list(cmd: commands.ListTask,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        filter_options = utils.FilterOptions.from_kwargs(**context)
        with bus.uow as uow:
            if filter_options:
                tasks = uow.tasks.get_by_conditions(
                    entities.task.Task,
                    filter_options.get_only_set_attributes())
            else:
                tasks = uow.tasks.list(entities.task.Task)
            if tasks:
                print_options = printer.PrintOptions.from_kwargs(**context)
                bus.printer.console.print_task(tasks, print_options)

    @register(cmd=commands.ShowTask)
    def show(cmd: commands.ShowTask,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            task = uow.tasks.get_by_id(entities.task.Task, cmd.id)
            bus.printer.tui.print_task(task, bus)


class TaskEventHandlers:

    @register(event=events.TaskCreated)
    def created(event: events.TaskCreated,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        # TODO: Decide what to do here
        ...

    @register(event=events.TaskCompleted)
    def completed(event: events.TaskCompleted,
                  bus: "messagebus.MessageBus",
                  context: dict = {}) -> None:
        with bus.uow as uow:
            task_event = entities.event_history.TaskEvent(**asdict(event))
            uow.tasks.add(task_event)
            uow.tasks.update(entities.task.Task, event.id,
                             {"modification_date": datetime.now()})
            uow.commit()

    @register(event=events.TaskUpdated)
    def updated(event: events.TaskUpdated,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            task_event = entities.event_history.TaskEvent(**asdict(event))
            uow.tasks.add(task_event)
            uow.tasks.update(entities.task.Task, event.task,
                             {"modification_date": datetime.now()})
            uow.commit()
            logging.debug(f"Task updated, context {event}")

    @register(event=events.TaskDeleted)
    def deleted(event: events.TaskDeleted,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            task_event = entities.event_history.TaskEvent(task_id=event.id,
                                                          type="STATUS",
                                                          old_value="",
                                                          new_value="DELETED")
            uow.tasks.add(task_event)
            uow.tasks.update(entities.task.Task, event.d,
                             {"modification_date": datetime.now()})
            uow.commit()

    @register(event=events.TaskCommentAdded)
    def comment_added(event: events.TaskCommentAdded,
                      bus: "messagebus.MessageBus",
                      context: dict = {}) -> None:
        TaskCommandHandlers.comment(cmd=commands.CommentTask(**asdict(event)),
                                    bus=bus,
                                    context=context)
        with bus.uow as uow:
            uow.tasks.update(entities.task.Task, event.id,
                             {"modification_date": datetime.now()})

    @register(event=events.TaskCollaboratorAdded)
    def collaborator_added(event: events.TaskCollaboratorAdded,
                           bus: "messagebus.MessageBus",
                           context: dict = {}) -> None:
        TaskCommandHandlers.collaborate(
            cmd=commands.CollaborateTask(**asdict(event)),
            bus=bus,
            context=context)

    @register(event=events.TaskHoursSubmitted)
    def hours_submitted(event: events.TaskHoursSubmitted,
                        bus: "messagebus.MessageBus",
                        context: dict = {}) -> None:
        TaskCommandHandlers.track(cmd=commands.TrackTask(id=event.id,
                                                         hours=event.hours),
                                  bus=bus,
                                  context=context)

    @register(event=events.TaskAddedToEpic)
    def added_to_epic(event: events.TaskAddedToEpic,
                      bus: "messagebus.MessageBus",
                      context: dict = {}) -> None:
        TaskCommandHandlers.add(cmd=commands.AddTask(id=event.id,
                                                     epic=event.epic_id),
                                bus=bus,
                                context=context)

    @register(event=events.TaskAddedToSprint)
    def added_to_sprint(event: events.TaskAddedToSprint,
                        bus: "messagebus.MessageBus",
                        context: dict = {}) -> None:
        TaskCommandHandlers.add(cmd=commands.AddTask(id=event.id,
                                                     sprint=event.sprint_id),
                                bus=bus,
                                context=context)

    @register(event=events.TaskAddedToStory)
    def added_to_story(event: events.TaskAddedToStory,
                       bus: "messagebus.MessageBus",
                       context: dict = {}) -> None:
        TaskCommandHandlers.add(cmd=commands.AddTask(id=event.id,
                                                     story=event.story_id),
                                bus=bus,
                                context=context)

    @register(event=events.TaskSynced)
    def synced(event: events.TaskSynced,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if synced_task := uow.tasks.get_by_conditions(
                    asana.AsanaTask, {"asana_task_id": event.asana_task_id}):
                uow.tasks.update(asana.AsanaTask, synced_task[0].id,
                                 {"sync_date": event.sync_date})
            else:
                uow.tasks.add(asana.AsanaTask(**asdict(event)))
            uow.commit()


class ProjectCommandHandlers:

    @register(cmd=commands.CreateProject)
    def create(cmd: commands.CreateProject,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                entities.project.Project, commands.CreateProject)
        project_id = None
        with bus.uow as uow:
            if not (existing_project := uow.tasks.get(entities.project.Project,
                                                      cmd.name)):
                cmd = convert_workspace(cmd, bus)
                new_project = entities.project.Project(**asdict(cmd))
                uow.tasks.add(new_project)
                uow.flush()
                project_id = int(new_project.id)
                new_event = events.ProjectCreated(project_id)
                uow.published_messages.append(new_event)
                uow.commit()
                bus.printer.console.print_new_object(new_project)
                bus.publisher.publish("Topic", new_event)
            else:
                logging.warning(f"Project {cmd.name} already exists")
                project_id = existing_project.id
            return project_id

    @register(cmd=commands.UpdateProject)
    def update(cmd: commands.UpdateProject,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            if not cmd:
                cmd, context = templates.create_command_from_editor(
                    project, commands.UpdateProject)
                cmd.id = project.id
            cmd.inject(bus.config)
            cmd = convert_workspace(cmd, bus)
            uow.tasks.update(entities.project.Project, project.id,
                             cmd.get_only_set_attributes())
            uow.commit()
            ProjectCommandHandlers._process_extra_args(project.id, context,
                                                       uow)

    @register(cmd=commands.CompleteProject)
    def complete(cmd: commands.CompleteProject,
                 bus: "messagebus.MessageBus",
                 context: dict = {}) -> None:
        with bus.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            uow.tasks.update(entities.project.Project, project.id,
                             {"status": "COMPLETED"})
            uow.commit()
            uow.published_messages.append(events.ProjectCompleted(project.id))
            ProjectCommandHandlers._process_extra_args(project.id, context,
                                                       uow)
            bus.publisher.publish("Topic", events.ProjectCompleted(project.id))

    @register(cmd=commands.DeleteProject)
    def delete(cmd: commands.DeleteProject,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            uow.tasks.update(entities.project.Project, project.id,
                             {"status": "DELETED"})
            uow.commit()
            uow.published_messages.append(events.ProjectDeleted(project.id))
            ProjectCommandHandlers._process_extra_args(project.id, context,
                                                       uow)
            bus.publisher.publish("Topic", events.ProjectDeleted(project.id))

    @register(cmd=commands.CommentProject)
    def comment(cmd: commands.CommentProject,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            uow.tasks.add(
                entities.commentary.ProjectCommentary(id=project.id,
                                                      text=cmd.text))
            uow.commit()

    @register(cmd=commands.TagProject)
    def tag(cmd: commands.TagProject,
            bus: "messagebus.MessageBus",
            context: dict = {}) -> None:
        with bus.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            # TODO: Extract into it's own method
            if not (existing_tag := uow.tasks.list(entities.tag.BaseTag,
                                                   {"text": cmd.tag})):
                new_tag = entities.tag.BaseTag(text=cmd.tag)
                uow.tasks.add(new_tag)
                uow.flush()
                tag_id = new_tag.id
            else:
                tag_id = existing_tag[0].id
            if not uow.tasks.list(entities.tag.ProjectTag, {
                    "project": project.id,
                    "tag": tag_id
            }):
                uow.tasks.add(
                    entities.tag.ProjectTag(id=project.id, tag_id=tag_id))
                uow.commit()

    @register(cmd=commands.CollaborateProject)
    def collaborate(cmd: commands.CollaborateProject,
                    bus: "messagebus.MessageBus",
                    context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_project := uow.tasks.get_by_id(
                    entities.project.Project, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Project id {cmd.id} is not found")
            if not (existing_user := uow.tasks.list(
                    entities.user.User, {"name": cmd.collaborator})):
                new_user = entities.user.User(name=cmd.collaborator)
                uow.tasks.add(new_user)
                uow.flush()
                user_id = new_user.id
            else:
                user_id = existing_user[0].id
            if not uow.tasks.list(entities.collaborator.ProjectCollaborator, {
                    "project": cmd.id,
                    "collaborator": user_id
            }):
                uow.tasks.add(
                    entities.collaborator.ProjectCollaborator(
                        id=cmd.id, collaborator_id=user_id))
                uow.commit()

    @register(cmd=commands.GetProject)
    def get(cmd: commands.GetProject,
            bus: "messagebus.MessageBus",
            context: dict = {}) -> list[entities.project.Project]:
        with bus.uow as uow:
            if project_id := cmd.id:
                return [ProjectCommandHandlers._validate_project(cmd.id, uow)]
            else:
                return uow.tasks.list(entities.project.Project)

    @register(cmd=commands.ShowProject)
    def show(cmd: commands.ShowProject,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            bus.printer.tui.print_project(project, bus)

    @register(cmd=commands.ListProject)
    def list(cmd: commands.ListProject,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            filter_options = utils.FilterOptions.from_kwargs(**context)
            with bus.uow as uow:
                if filter_options:
                    projects = uow.tasks.get_by_conditions(
                        entities.project.Project,
                        filter_options.get_only_set_attributes())
                else:
                    projects = uow.tasks.list(entities.project.Project)
                bus.printer.console.print_project(
                    projects, printer.PrintOptions.from_kwargs(**context))

    @register(cmd=commands.SyncProject)
    def sync(cmd: commands.SyncProject,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            if cmd.id:
                project = ProjectCommandHandlers._validate_project(cmd.id, uow)
                ProjectCommandHandlers._sync_project(uow, project)
            else:
                projects = uow.tasks.list(entities.project.Project)
                for project in projects:
                    ProjectCommandHandlers._sync_project(uow, project)

    @register(cmd=commands.ConnectProject)
    def connect(cmd: commands.ConnectProject,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            project = ProjectCommandHandlers._validate_project(cmd.id, uow)
            if not (asana_project := uow.tasks.get_by_id(
                    asana.AsanaProject, project.id)):
                asana_project = asana.AsanaProject(
                    id=project.id,
                    asana_project_id=cmd.external_project,
                    sync_date=None)
                uow.tasks.add(asana_project)
                uow.commit()

    def _validate_project(id: str | int, uow) -> entities.project.Project:
        if isinstance(id, int) or id.isnumeric():
            if not (existing_project := uow.tasks.get_by_id(
                    entities.project.Project, id)):
                raise exceptions.EntityNotFound(
                    f"Project id {id} is not found")
        else:
            if not (existing_project := uow.tasks.get(entities.project.Project,
                                                      id)):

                raise exceptions.EntityNotFound(
                    f"Project id {id} is not found")
        return existing_project

    def _sync_project(uow, project: entities.project.Project):
        asana_project = uow.tasks.get_by_id(asana.AsanaProject, project.id)
        if not asana_project:
            return
        asana_project_id = asana_project.asana_project_id
        if last_sync_date := asana_project.sync_date:
            tasks = [
                task for task in project.tasks
                if task.creation_date > last_sync_date or (
                    task.modification_date
                    and task.modification_date > last_sync_date)
            ]
        else:
            tasks = project.tasks
        if not tasks:
            return
        project_sync_date = datetime.now()
        synced_tasks = views.external_connectors_asana_tasks(
            uow.repo.session, project.id)
        mapped_external_users = views.external_connectors_asana_users(
            uow.repo.session)
        # TODO: Store default assign user and token in config
        configuration = asn.Configuration()
        configuration.access_token = os.getenv("ASANA_PERSONAL_ACCESS_TOKEN")
        asana_client = asn.ApiClient(configuration)
        asana_migrator = asana.AsanaMigrator(asana_client)
        asana_migrator.load_task_statuses(asana_project_id)
        for i, task in enumerate(tasks):
            if not task.sync:
                continue
            sync_info = synced_tasks.get(task.id)
            if asana_task_id := asana_migrator.migrate_task(
                    asana_project_id, task, sync_info, mapped_external_users):
                uow.published_messages.append(
                    events.TaskSynced(id=task.id,
                                      project=project.id,
                                      asana_task_id=asana_task_id,
                                      sync_date=datetime.now()))
        uow.published_messages.append(
            events.ProjectSynced(project.id, asana_project_id,
                                 project_sync_date))

    def _process_extra_args(id, context, uow):
        if tags := context.get("tags"):
            for tag in tags.split(","):
                uow.published_messages.append(commands.TagProject(id=id,
                                                                tag=tag))
        if collaborators := context.get("collaborators"):
            for collaborator_name in collaborators.split(","):
                uow.published_messages.append(
                    commands.CollaborateProject(
                        id=id, collaborator=collaborator_name))
        if comment := context.get("comment"):
            uow.published_messages.append(
                commands.CommentProject(id=id, text=comment))


class ProjectEventHandlers:

    @register(event=events.ProjectCreated)
    def created(event: events.ProjectCreated,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        # TODO: Decide what to do here
        ...

    @register(event=events.ProjectCommented)
    def commented(event: events.ProjectCommented,
                  bus: "messagebus.MessageBus",
                  context: dict = {}) -> None:
        ProjectCommandHandlers.comment(
            cmd=commands.CommentProject(**asdict(event)), bus=bus)

    @register(event=events.ProjectSynced)
    def synced(event: events.ProjectSynced,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if synced_project := uow.tasks.get_by_conditions(
                    asana.AsanaProject,
                {"asana_project_id": event.asana_project_id}):
                uow.tasks.update(asana.AsanaProject, synced_project[0].id,
                                 {"sync_date": event.sync_date})
            else:
                uow.tasks.add(asana.AsanaProject(**asdict(event)))
            uow.commit()


class EpicCommandHandlers:

    @register(cmd=commands.CreateEpic)
    def create(cmd: commands.CreateEpic,
               bus: "messagebus.MessageBus",
               context: dict = {}):
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                entities.epic.Epic, type(cmd))
        with bus.uow as uow:
            cmd = convert_project(cmd, bus)
            new_epic = entities.epic.Epic(**asdict(cmd))
            uow.tasks.add(new_epic)
            uow.commit()
            new_epic_id = new_epic.id
            bus.printer.console.print_new_object(new_epic)
            return new_epic_id

    @register(cmd=commands.UpdateEpic)
    def update(cmd: commands.UpdateEpic,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_epic := uow.tasks.get_by_id(
                    entities.epic.Epic, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Epic id {cmd.id} is not found")
            if not cmd:
                cmd, context = templates.create_command_from_editor(
                    existing_epic, commands.UpdateEpic)
            cmd.inject(bus.config)
            uow.tasks.update(entities.epic.Epic, cmd.id,
                             cmd.get_only_set_attributes())
            uow.commit()

    @register(cmd=commands.CompleteEpic)
    def complete(cmd: commands.CompleteEpic,
                 bus: "messagebus.MessageBus",
                 context: dict = {}) -> None:
        with bus.uow as uow:
            uow.tasks.update(entities.project.Epic, cmd.id,
                             {"status": "COMPLETED"})
            uow.commit()
            uow.published_messages.append(events.EpicCompleted(cmd.id))
            EpicCommandHandlers._process_extra_args(cmd.id, context, uow)
        bus.publisher.publish("Topic", events.EpicCompleted(cmd.id))

    @register(cmd=commands.DeleteEpic)
    def delete(cmd: commands.DeleteEpic,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            uow.tasks.update(entities.epic.Epic, cmd.id, {"status": "DELETED"})
            uow.commit()
            uow.published_messages.append(events.EpicDeleted(cmd.id))
            EpicCommandHandlers._process_extra_args(cmd.id, context, uow)
        bus.publisher.publish("Topic", events.EpicDeleted(cmd.id))

    @register(cmd=commands.CommentEpic)
    def comment(cmd: commands.CommentEpic,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            uow.tasks.add(
                entities.commentary.EpicCommentary(id=cmd.id, text=cmd.text))
            uow.commit()

    @register(cmd=commands.AddEpic)
    def add(cmd: commands.AddEpic,
            bus: "messagebus.MessageBus",
            context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_epic := uow.tasks.get_by_id(
                    entities.epic.Epic, cmd.id)):
                raise exceptions.EntityNotFound(f"Epic {cmd.id} is not found")
            if not uow.tasks.get_by_id(entities.sprint.Sprint, cmd.sprint):
                raise exceptions.EntityNotFound(
                    f"Sprint {cmd.sprint} is not found")
            for epic_task in existing_epic.tasks:
                task = epic_task.tasks
                if task.status.name not in ("DONE", "DELETED"):
                    uow.published_messages.append(
                        commands.AddTask(id=task.id, sprint=cmd.sprint))

    @register(cmd=commands.ListEpic)
    def list(cmd: commands.ListEpic,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            if epics := uow.tasks.list(entities.epic.Epic):
                bus.printer.console.print_composite(
                    epics, printer.PrintOptions.from_kwargs(**context), "epic")

    @register(cmd=commands.ShowEpic)
    def show(cmd: commands.ShowEpic,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            epic = uow.tasks.get_by_id(entities.epic.Epic, cmd.id)
            bus.printer.tui.print_epic(epic, bus)

    def _process_extra_args(id, context, uow):
        if comment := context.get("comment"):
            uow.published_messages.append(
                commands.CommentEpic(id=id, text=comment))


class EpicEventHandlers:
    ...


class StoryCommandHandlers:

    @register(cmd=commands.CreateStory)
    def create(cmd: commands.CreateStory,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> int:
        if not cmd.name:
            cmd, context = templates.create_command_from_editor(
                entities.story.Story, type(cmd))
        with bus.uow as uow:
            cmd = convert_project(cmd, bus)
            new_story = entities.story.Story(**asdict(cmd))
            uow.tasks.add(new_story)
            uow.commit()
            new_story_id = new_story.id
            bus.printer.console.print_new_object(new_story)
            return new_story_id

    @register(cmd=commands.UpdateStory)
    def update(cmd: commands.UpdateStory,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_story := uow.tasks.get_by_id(
                    entities.story.Story, cmd.id)):
                raise exceptions.EntityNotFound(
                    f"Story id {cmd.id} is not found")
            if not cmd:
                cmd, context = templates.create_command_from_editor(
                    existing_story, commands.UpdateStory)
            cmd.inject(bus.config)
            uow.tasks.update(entities.story.Story, cmd.id,
                             cmd.get_only_set_attributes())
            uow.commit()

    @register(cmd=commands.CompleteStory)
    def complete(cmd: commands.CompleteStory,
                 bus: "messagebus.MessageBus",
                 context: dict = {}) -> None:
        with bus.uow as uow:
            uow.tasks.update(entities.project.Story, cmd.id,
                             {"status": "COMPLETED"})
            uow.commit()
            uow.published_messages.append(events.StoryDeleted(cmd.id))
            StoryCommandHandlers._process_extra_args(cmd.id, context, uow)
        bus.publisher.publish("Topic", events.StoryCompleted(cmd.id))

    @register(cmd=commands.DeleteStory)
    def delete(cmd: commands.DeleteStory,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            uow.tasks.update(entities.story.Story, cmd.id,
                             {"status": "DELETED"})
            uow.commit()
            uow.published_messages.append(events.StoryDeleted(cmd.id))
            StoryCommandHandlers._process_extra_args(cmd.id, context, uow)
        bus.publisher.publish("Topic", events.StoryDeleted(cmd.id))

    @register(cmd=commands.CommentStory)
    def comment(cmd: commands.CommentStory,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            uow.tasks.add(
                entities.commentary.StoryCommentary(id=cmd.id, text=cmd.text))
            uow.commit()

    @register(cmd=commands.AddStory)
    def add(cmd: commands.AddStory,
            bus: "messagebus.MessageBus",
            context: dict = {}) -> None:
        with bus.uow as uow:
            if not (existing_story := uow.tasks.get_by_id(
                    entities.story.Story, cmd.id)):
                raise exceptions.EntityNotFound(f"Story {cmd.id} is not found")
            if not uow.tasks.get_by_id(entities.sprint.Sprint, cmd.sprint):
                raise exceptions.EntityNotFound(
                    f"Sprint {cmd.sprint} is not found")
            for story_task in existing_story.tasks:
                task = story_task.tasks
                if task.status.name not in ("DONE", "DELETED"):
                    uow.published_messages.append(
                        commands.AddTask(id=task.id, sprint=cmd.sprint))

    @register(cmd=commands.ListStory)
    def list(cmd: commands.ListStory,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            if storys := uow.tasks.list(entities.story.Story):
                bus.printer.console.print_composite(
                    storys, printer.PrintOptions.from_kwargs(**context),
                    "story")

    @register(cmd=commands.ShowStory)
    def show(cmd: commands.ShowStory,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            story = uow.tasks.get_by_id(entities.story.Story, cmd.id)
            bus.printer.tui.print_story(story, bus)

    def _process_extra_args(id, context, uow):
        if comment := context.get("comment"):
            uow.published_messages.append(
                commands.CommentStory(id=id, text=comment))


class WorkspaceCommandHandlers:

    @register(cmd=commands.CreateWorkspace)
    def create(cmd: commands.CreateWorkspace,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        if not cmd:
            cmd, context = templates.create_command_from_editor(
                entities.sprint.Sprint, type(cmd))
        with bus.uow as uow:
            if not uow.tasks.get(entities.workspace.Workspace, cmd.name):
                new_workspace = entities.workspace.Workspace(**asdict(cmd))
                uow.tasks.add(new_workspace)
                uow.commit()
                bus.printer.console.print_new_object(new_workspace)
            else:
                logging.warning(f"Workspace {cmd.name} already exists")


class TagCommandHandlers:

    @register(cmd=commands.CreateTag)
    def create(cmd: commands.CreateTag,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if not uow.tasks.get_by_conditions(entities.tag.BaseTag,
                                               {"text": cmd.text}):
                new_tag = entities.tag.BaseTag(**asdict(cmd))
                uow.tasks.add(new_tag)
                uow.commit()
                bus.printer.console.print_new_object(new_tag, {})
                return cmd.text
            else:
                logging.warning(f"Tag {cmd.text} already exists")

    @register(cmd=commands.DeleteTag)
    def delete(cmd: commands.DeleteTag,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if tags := uow.tasks.get_by_conditions(entities.tag.BaseTag,
                                                   {"text": cmd.text}):
                uow.tasks.delete(entities.tag.BaseTag, tags[0].id)
                uow.commit()
                return cmd.text
            else:
                logging.warning(f"Tag {cmd.text} does not exists")

    @register(cmd=commands.ListTag)
    def list(cmd: commands.ListTag,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            if tags := uow.tasks.list(entities.tag.BaseTag):
                bus.printer.console.print_tag(tags)


class UserCommandHandlers:

    @register(cmd=commands.CreateUser)
    def create(cmd: commands.CreateUser,
               bus: "messagebus.MessageBus",
               context: dict = {}) -> None:
        with bus.uow as uow:
            if not uow.tasks.get(entities.user.User, cmd.name):
                new_user = entities.user.User(**asdict(cmd))
                uow.tasks.add(new_user)
                uow.commit()
                bus.printer.console.print_new_object(new_user)
            else:
                logging.warning(f"User {cmd.name} already exists")

    @register(cmd=commands.ListUser)
    def list(cmd: commands.ListUser,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            if users := uow.tasks.list(entities.user.User):
                bus.printer.console.print_user(users)

    @register(cmd=commands.ConnectUser)
    def connect(cmd: commands.ConnectUser,
                bus: "messagebus.MessageBus",
                context: dict = {}) -> None:
        with bus.uow as uow:
            user = UserCommandHandlers._validate_user(cmd.id, uow)
            if not (asana_user := uow.tasks.get_by_id(asana.AsanaUser,
                                                      user.id)):
                asana_user = asana.AsanaUser(id=user.id,
                                             asana_user_id=cmd.external_user)
                uow.tasks.add(asana_user)
                uow.commit()

    def _validate_user(id: str | int, uow) -> entities.user.User:
        if isinstance(id, int) or id.isnumeric():
            if not (existing_user := uow.tasks.get_by_id(
                    entities.user.User, id)):
                raise exceptions.EntityNotFound(f"User {id} is not found")
        else:
            if not (existing_user := uow.tasks.get(entities.user.User, id)):

                raise exceptions.EntityNotFound(f"User {id} is not found")
        return existing_user


class NoteCommandHandlers:

    @register(cmd=commands.ShowNote)
    def show(cmd: commands.ShowNote,
             bus: "messagebus.MessageBus",
             context: dict = {}) -> None:
        with bus.uow as uow:
            note_type = get_note_type(context)
            if note := uow.tasks.get_by_id(note_type, cmd.id):
                bus.printer.tui.show_note(note)


def convert_project(cmd: commands.Command,
                    bus: "messagebus.MessageBus",
                    context: dict = {}) -> Type[commands.Command]:
    if not (project_name := cmd.project):
        cmd.project = None
        return cmd
    if isinstance(project_name, int):
        return cmd
    if project_name.isnumeric():
        cmd.project = int(project_name)
        return cmd
    if not (existing_project := bus.uow.tasks.get(entities.project.Project,
                                                  project_name)):
        answer = Confirm.ask(
            f"Creating new project: {project_name}. Do you want to continue?")
        if not answer:
            project_name = Prompt.ask("Provide a project name: ")
            answer = Confirm.ask(
                f"Creating new project: {project_name}. Do you want to continue?"
            )
        project_id = ProjectCommandHandlers.create(
            cmd=commands.CreateProject(name=project_name),
            bus=bus,
            context=context)
        cmd.project = project_id
    else:
        cmd.project = int(existing_project.id)
    return cmd


def convert_workspace(cmd: commands.Command,
                      bus: "messagebus.MessageBus",
                      context: dict = {}) -> Type[commands.Command]:
    try:
        cmd.workspace = int(cmd.workspace)
        return cmd
    except ValueError:
        ...
    if not (workspace := cmd.workspace):
        workspace = bus.config.get("workspace")
    if not (existing_workspace := bus.uow.tasks.get(
            entities.workspace.Workspace, workspace)):
        print(f"Creating new workspace: {workspace}. "
              "Do you want to continue (Y/n)?")
        answer = input()
        while answer.lower() != "y":
            print("Provide workspace name: ")
            project_name = input()
            print(f"Creating new workspace: {workspace}. "
                  "Do you want to continue (Y/n)?")
            answer = input()
        workspace_id = WorkspaceCommandHandlers.create(
            cmd=commands.CreateWorkspace(name=workspace),
            bus=bus,
            context=context)
        cmd.workspace = workspace_id
    else:
        cmd.workspace = int(existing_workspace.id)
    return cmd


def convert_user(cmd: commands.Command,
                 bus: "messagebus.MessageBus",
                 user_type: str = "created_by",
                 context: dict = {}) -> Type[commands.Command]:
    if not (user_value := getattr(cmd, user_type)):
        return cmd
    try:
        setattr(cmd, user_type, int(user_value))
        return cmd
    except ValueError:
        ...
    if not (existing_user := bus.uow.tasks.get(entities.user.User,
                                               user_value)):
        user_id = UserCommandHandlers.create(
            cmd=commands.CreateUser(name=user_value), bus=bus, context=context)
        setattr(cmd, user_type, user_id)
    else:
        setattr(cmd, user_type, int(existing_user.id))
    return cmd


def get_note_type(kwargs: dict[str, str]) -> entities.note.BaseNote:
    if "project" in kwargs:
        return entities.note.ProjectNote
    if "task" in kwargs:
        return entities.note.TaskNote
    if "sprint" in kwargs:
        return entities.note.SprintNote
    if "story" in kwargs:
        return entities.note.StoryNote
    if "epic" in kwargs:
        return entities.note.EpicNote
